#include "flib_cp_proc.h"
#include "flib_groups.h"
#include "flib_debug.h"
#include "flib_recovery.h"
#include "flib_logging_protocol.h"

void cp_process_notifications() {
	assert(is_cp_proc(get_initial_user_rank()));

	MPI_Status s;
	int cm_procs_done = 0;

	while(cm_procs_done < assigned_cm_procs_nr() ) {
		// first, we check whether we should initiate the non-blocking coordinated checkpoint 
		cp_manage_coordinated_checkpoint(&cm_procs_done);

		// then, we check for other incoming messages 
		MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &((*desc)->awaiting_requests), &s);

		if((*desc)->awaiting_requests == TRUE) {
			(*desc)->awaiting_requests = FALSE;

			int sender = s.MPI_SOURCE;
			int tag = s.MPI_TAG;

			if(tag == CP_MASTER_COORDINATED_CHECKPOINT_REQUEST) {
				handle_cp_master_coordinated_checkpoint_request(&cm_procs_done);
			}
			else if(tag == CM_COORDINATED_CHECKPOINT || tag == CM_FAKED_COORDINATED_CHECKPOINT) {
				handle_simple_bsp_coordinated_checkpoint(FALSE);
			}
			else if(tag == CM_DEMAND_PUT_CHECKPOINT_REQUEST) {      
				handle_cm_demand_put_checkpoint_request();
			}
			else if(tag == CM_DEMAND_PUT_CHECKPOINT_DONE) {
				handle_cm_demand_put_checkpoint_done();
			}
			else if(tag == CM_DEMAND_GET_CHECKPOINT_REQUEST) {      
				handle_cm_demand_get_checkpoint_request();
			}
			else if(tag == CM_DEMAND_GET_CHECKPOINT_DONE) {
				handle_cm_demand_get_checkpoint_done();
			}
			else if(tag == CM_NECESSARY_CHECKPOINT_DONE) {
				handle_cm_necessary_checkpoint_done();
			}
			else if(tag == CM_GLOBAL_LOGS_CLEARED) {
				handle_cm_global_logs_cleared();
			}
			else if(tag == CM_PROC_ERROR) {
				handle_proc_error();
			}
			else if(tag == CM_PROC_GLOBAL_RECOVERY) {
				handle_proc_recovery();
			}
			else if(tag == CP_MASTER_GLOBAL_RECOVERY_REQUEST) {
				handle_cp_master_global_recovery_request();
			}
			else if(tag == CM_QUIT) {
				handle_quit(&cm_procs_done);
			}
			else {
				handle_unknown_request_cp(tag);
			}
		}
	}

#ifdef FLIB_DEBUG
	debug_message_gen_e( FALSE, stderr, "QQQ CP: exiting\n");
#endif
}

int is_our_proc(int rank, int communicator_type) {
	//TODO: make it more flexible
	assert(communicator_type == _FLIB_INITIAL_COMM);

	int* assigned_procs = assigned_cm_procs();
	int size = assigned_cm_procs_nr();
	int our_proc = FALSE;
	int i = 0;
	for(; i < size; i++) {
		if(rank == assigned_procs[i]) {
			our_proc = TRUE;
			break;
		}
	}

	free(assigned_procs);
	return our_proc;
}

void handle_simple_bsp_coordinated_checkpoint(int faked) {
	assert((*desc)->coordinated_protocol_type == FLIB_BSP_SIMPLE || (*desc)->coordinated_protocol_type == FLIB_ORDERED_CHECKPOINT || faked == TRUE);

	MPI_Status s;
	uint64_t* checkpoint_data = &((uint64_t*)(*desc)->cp_data)[(*desc)->data_offset];
	uint64_t* xored_data = &((uint64_t*)(*desc)->cp_data)[(*desc)->xor_data_offset];

	int app_data_size = (*desc)->app_data_size;
	int cp_data_unit = (*desc)->cp_data_disp_unit;

	MPI_Recv(checkpoint_data, app_data_size / cp_data_unit, MPI_UINT64_T, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);

	assert(s.MPI_TAG == CM_COORDINATED_CHECKPOINT || s.MPI_TAG == CM_FAKED_COORDINATED_CHECKPOINT);
	assert(is_cm_proc(s.MPI_SOURCE));
	assert(is_our_proc(s.MPI_SOURCE, _FLIB_INITIAL_COMM) == TRUE);

	if((*cp_desc)->received_coordinated_checkpoints_nr == 0) {
		assert((*desc)->coordinated_checkpoint_started == FALSE);
		(*desc)->coordinated_checkpoint_started = TRUE;
	}

	(*cp_desc)->received_coordinated_checkpoints_nr++;
	if((*cp_desc)->received_coordinated_checkpoints_nr < ((*desc)->group_size - 1) && ((*cp_desc)->received_coordinated_checkpoint)[s.MPI_SOURCE] == TRUE) {
		/*TODO: do it correctly, probably with a queue of messages*/
		debug_message_gen_e(TRUE, stderr, "############## Coordinated checkpoins start to overlap, consider extending the checkpointing interval");
	}

	if((*cp_desc)->received_coordinated_checkpoints_nr == 1) {
		/* upon the first message, we just copy the data. */
		/* after the first message, we perform the actual xoring*/

#ifdef FLIB_COORDINATED_DEBUG
		debug_message_gen(FALSE, stderr, "&&& CP: received the first coordinated Checkpoint (from CM %i)\n", s.MPI_SOURCE);
#endif

		memcpy(xored_data, checkpoint_data, app_data_size);
		((*cp_desc)->received_coordinated_checkpoint)[s.MPI_SOURCE] = TRUE;
	}
	else if((*cp_desc)->received_coordinated_checkpoints_nr > 1 && (*cp_desc)->received_coordinated_checkpoints_nr < ((*desc)->group_size - 1)) {
		calculate_xor(checkpoint_data, xored_data, app_data_size);
		((*cp_desc)->received_coordinated_checkpoint)[s.MPI_SOURCE] = TRUE;

#ifdef FLIB_COORDINATED_DEBUG
		debug_message_gen(FALSE, stderr, "&&& CP: received coordinated Checkpoint nr %i (from CM %i)\n", (*cp_desc)->received_coordinated_checkpoints_nr, s.MPI_SOURCE);
#endif

	}
	else if((*cp_desc)->received_coordinated_checkpoints_nr == ((*desc)->group_size - 1)) {
		int i = 0;
		for(; i < (*desc)->initial_user_size; i++) {
			((*cp_desc)->received_coordinated_checkpoint)[i] = FALSE;
		}

		calculate_xor(checkpoint_data, xored_data, app_data_size);

#ifdef FLIB_COORDINATED_DEBUG
		debug_message_gen(FALSE, stderr, "&&& CP: received the last coordinated Checkpoint (from CM %i)\n", s.MPI_SOURCE);
#endif

		(*cp_desc)->received_coordinated_checkpoints_nr = 0;
		(*desc)->coordinated_checkpoint_started = FALSE;
	}
	else {
		debug_message_gen(TRUE, stderr, "############# CP: received the improper coordinated Checkpoint message (from CM %i)\n", s.MPI_SOURCE);
	}

#ifdef FLIB_COORDINATED_DEBUG
	debug_message_gen(FALSE, stderr, "&&& CP: another CM (%i) processed in BSP SIMPLE \n", (*cp_desc)->received_coordinated_checkpoints_nr);
#endif
}

void cp_manage_coordinated_checkpoint(int* cm_procs_done) {
	if(!is_first_cp_process() || (*desc)->coordinated_protocol_type == FLIB_BSP_SIMPLE || (*desc)->coordinated_protocol_type == FLIB_NO_COORDINATED_SCHEME ||
			(*desc)->coordinated_protocol_type == FLIB_ORDERED_CHECKPOINT) {
		return;
	}
	double elapsed_time = MPI_Wtime();
	double prev_cp_time = (*desc)->prev_cp_time;
	double daly_interval = (*desc)->daly_interval;
	double current_interval = elapsed_time - prev_cp_time;
	assert(current_interval > 0);

	if(current_interval > daly_interval) {
#ifdef FLIB_COORDINATED_DEBUG
		debug_message_gen_e( FALSE, stderr, "&&& Beginning coordinated checkpoint\n");
#endif
		(*desc)->coordinated_checkpoint_started = TRUE;

		/* first, we send the request to other CP processes */
		send_cp_master_notification(CP_MASTER_COORDINATED_CHECKPOINT_REQUEST);

		/* then,we send the actual checkpoint request to our CM procs */
		send_cp_notification_to_assigned_cms(CP_COORDINATED_CHECKPOINT_REQUEST);	

		/* then we wait for the checkpoints to arrive */
		wait_for_coordinated_checkpoints(cm_procs_done);

#ifdef FLIB_COORDINATED_DEBUG
		debug_message_gen_e( FALSE, stderr, "&&& Checkpoint DONE, waiting for others\n");
#endif		

		//MPI_Barrier((*desc)->initial_user_comm);
		MPI_Barrier(*(*desc)->flib_level_comm);
		(*desc)->coordinated_checkpoint_started = FALSE;

		(*desc)->prev_cp_time = MPI_Wtime();

#ifdef FLIB_COORDINATED_DEBUG
		debug_message_gen_e( FALSE, stderr, "&&& Coordinated checkpoint barrier passed\n");
#endif
	}
}

void send_cp_master_notification(int tag) {
	/* TODO: more efficient sending (collectives maybe?) */
	int i = 0;
	int cp_procs = (*desc)->cp_size;
	int* cp_procs_ranks = get_cp_procs_ranks();

	for(; i < cp_procs; i++) {
		assert(is_cp_proc(cp_procs_ranks[i]));

		if(cp_procs_ranks[i] != (*desc)->initial_user_rank) {
			assert(i != (*desc)->flib_level_rank);
			MPI_Send(NULL, 0, MPI_CHAR, cp_procs_ranks[i], tag, (*desc)->initial_user_comm);
		}
	}
	free(cp_procs_ranks);

	if(tag == CP_MASTER_COORDINATED_CHECKPOINT_REQUEST) {
		debug_message_gen_e( FALSE, stderr, "&&& sent coordinated checkpoint request to other CP procs\n");
	}
	else if(tag == CP_MASTER_GLOBAL_RECOVERY_REQUEST) {
		debug_message_gen_e( FALSE, stderr, "^^^ sent global recovery request to other CP procs\n");
	}
}

void send_cp_notification_to_assigned_cms(int tag) {
	int* assigned_procs = assigned_cm_procs();
	int i = 0;
	for(; i < assigned_cm_procs_nr(); i++) {
		MPI_Send(NULL, 0, MPI_CHAR, assigned_procs[i], tag, (*desc)->initial_user_comm);
	}

#ifdef FLIB_COORDINATED_DEBUG
	if(tag == CP_COORDINATED_CHECKPOINT_REQUEST) {
		debug_message_gen_e( FALSE, stderr, "&&& sent coordinated checkpoint request to assigned CM procs\n");
	}
#endif
#ifdef FLIB_GLOBAL_RECOVERY_DEBUG
	if(tag == CP_GLOBAL_RECOVERY_REQUEST) {
		debug_message_gen_e( FALSE, stderr, "^^^ sent global recovery request to assigned CM procs\n");
	}
#endif

	free(assigned_procs);		
}		

void handle_cp_master_coordinated_checkpoint_request(int* cm_procs_done) {
	assert(cp_proc());
	assert(!is_first_cp_process());
	assert((*desc)->coordinated_protocol_type != FLIB_BSP_SIMPLE && (*desc)->coordinated_protocol_type != FLIB_NO_COORDINATED_SCHEME && 
			(*desc)->coordinated_protocol_type != FLIB_ORDERED_CHECKPOINT);

	MPI_Status s;
	MPI_Recv(NULL, 0, MPI_CHAR, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);

	assert(is_cp_proc(s.MPI_SOURCE));
	assert(first_cp_proc(s.MPI_SOURCE));
	assert(s.MPI_TAG == CP_MASTER_COORDINATED_CHECKPOINT_REQUEST);

	(*desc)->coordinated_checkpoint_started = TRUE;

#ifdef FLIB_COORDINATED_DEBUG	
	debug_message_gen_e( FALSE, stderr, "&&& CP: received CP COORDINATED CHECKPOINT request from master \n");	
#endif

	send_cp_notification_to_assigned_cms(CP_COORDINATED_CHECKPOINT_REQUEST);
	wait_for_coordinated_checkpoints(cm_procs_done);

#ifdef FLIB_COORDINATED_DEBUG
	debug_message_gen_e( FALSE, stderr, "&&& CP: Checkpoint DONE, waiting for other CPs \n");
#endif

	//MPI_Barrier((*desc)->initial_user_comm);
	MPI_Barrier(*(*desc)->flib_level_comm);

	(*desc)->coordinated_checkpoint_started = FALSE;

#ifdef FLIB_COORDINATED_DEBUG
	debug_message_gen_e( FALSE, stderr, "&&& CP: Barrier passed \n");
#endif
}

void wait_for_coordinated_checkpoints(int* cm_procs_done) {
	MPI_Status s;
	int cm_procs_checkpointed = 0;
	int requests = FALSE;
	int* assigned_procs = assigned_cm_procs();
	int assigned_procs_nr = assigned_cm_procs_nr();

	uint64_t* checkpoint_data = &((uint64_t*)(*desc)->cp_data)[(*desc)->data_offset];
	uint64_t* xored_data = &((uint64_t*)(*desc)->cp_data)[(*desc)->xor_data_offset];

	int app_data_size = (*desc)->app_data_size;
	int cp_data_unit = (*desc)->cp_data_disp_unit;

	/* CP proc processes the requests until all assigned CM procs send the confirmation */
	while(cm_procs_checkpointed < assigned_procs_nr && *cm_procs_done < assigned_procs_nr) {
		MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &requests, &s);

		if(requests == TRUE) {
			requests = FALSE;
			int tag = s.MPI_TAG;
			if(tag != CM_COORDINATED_CHECKPOINT) {
				/* we skip any other checkpoint requests at this stage */
				if(tag == CM_GLOBAL_RECOVERY_CONFIRMATION || tag == CM_PROC_GLOBAL_RECOVERY || tag == CP_MASTER_GLOBAL_RECOVERY_REQUEST) {
					debug_message_gen_e( TRUE, stderr,"############ Sth went wrong, cannot perform a global recovery and do a global checkpoint at the same time \n");
				}
				else if(tag == CP_MASTER_COORDINATED_CHECKPOINT_REQUEST) {
					debug_message_gen_e( TRUE, stderr,"############ Sth went wrong, coordinated checkpoint is already in progress... \n");
				}
				else if(tag == CM_DEMAND_PUT_CHECKPOINT_REQUEST) {
					int faked_data;
					MPI_Recv(&faked_data, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, MPI_STATUS_IGNORE);
					debug_message_gen_e( FALSE, stderr,"############ Skipping a demand PUT checkpoint request \n");
				}
				else if(tag == CM_DEMAND_PUT_CHECKPOINT_DONE || tag == CM_NECESSARY_CHECKPOINT_DONE) {
					int faked_data[3];
					MPI_Recv(&faked_data[0], 3, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, MPI_STATUS_IGNORE);
					debug_message_gen_e( FALSE, stderr,"############ Skipping a demand PUT / necessary checkpoint done \n");
				}
				else if(tag == CM_PROC_ERROR) {
					int dead_rank;
					MPI_Recv(&dead_rank, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);
					debug_message_gen_e( TRUE, stderr,"############ Cannot handle a crash during a coordinated checkpoint, exiting... \n");
				}
				else if(tag == CM_QUIT) {
					debug_message_gen_e( TRUE, stderr,"############ CM_QUIT during coordinated checkpoint, TODO...\n");
				}
				else {
					handle_unknown_request_cp(tag);
				}
			}
			else {
				MPI_Recv(checkpoint_data, app_data_size / cp_data_unit, MPI_UINT64_T, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);

				assert(s.MPI_TAG == CM_COORDINATED_CHECKPOINT);
				assert(is_cm_proc(s.MPI_SOURCE));
				assert(is_our_proc( s.MPI_SOURCE, _FLIB_INITIAL_COMM) == TRUE);

				if(cm_procs_checkpointed == 0 ) {
					memcpy(xored_data, checkpoint_data, app_data_size);
				}
				else {
					calculate_xor(checkpoint_data, xored_data, app_data_size);
				}

				cm_procs_checkpointed++;
#ifdef FLIB_COORDINATED_DEBUG
				debug_message_gen( FALSE, stderr,"CP: another CM (%i) processed \n", cm_procs_checkpointed);
#endif
			}
		}
	}
	free(assigned_procs);
}

//#######################################################################################################################################
//PUT demand checkpoint requests
//#######################################################################################################################################

void handle_cm_demand_put_checkpoint_request() {
	if(!do_logging()) {
		debug_message_gen_e(TRUE, stderr,"########### Cannot receive cm demand PUT checkpoint request if no logging enables!\n");
	}
	MPI_Status s;
	int target_rank;

	MPI_Recv(&target_rank, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);
	assert(s.MPI_TAG == CM_DEMAND_PUT_CHECKPOINT_REQUEST);
	assert(is_cm_proc(s.MPI_SOURCE));	
	assert(is_our_proc(target_rank, _FLIB_INITIAL_COMM));

#ifdef FLIB_DEMAND_DEBUG
	debug_message_gen( FALSE, stderr, "*** CP: received demand PUT checkpoint request for CM [%i,IN] from CM [%i,IN]\n", target_rank, s.MPI_SOURCE);
#endif

	//we do not send multiple demand PUT checkpoint requests
	//if a message has already been sent, we just add the requesting proc to a local queue to let it know later 
	if(gdsl_queue_is_empty((*cp_desc)->queued_demand_checkpoint_requests[target_rank])) {
		MPI_Send(&s.MPI_SOURCE, 1, MPI_INT, target_rank, CP_DEMAND_PUT_CHECKPOINT_REQUEST, (*desc)->initial_user_comm);
	}

	int* next_source_proc = malloc(sizeof(int));
	*next_source_proc = s.MPI_SOURCE;
	gdsl_queue_insert( (*cp_desc)->queued_demand_checkpoint_requests[target_rank], (void*)next_source_proc );	//TODO: maybe change the name

#ifdef FLIB_DEMAND_DEBUG
	debug_message_gen( FALSE, stderr, "*** CP: sent demand PUT checkpoint request to CM [%i,IN]\n", target_rank);
#endif
}

void handle_cm_demand_put_checkpoint_done() {
	if(!do_logging()) {
		debug_message_gen_e(TRUE, stderr,"########### Cannot receive cm demand PUT checkpoint done if no logging enables!\n");
	}
	MPI_Status s;
	//int data[3];
	int target_rank;

	//this time 'target_rank' means the process that originally requested a demand checkpoint
	MPI_Recv(&target_rank, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);

	assert(is_cm_proc(s.MPI_SOURCE));
	assert(s.MPI_TAG == CM_DEMAND_PUT_CHECKPOINT_DONE);
	assert(is_our_proc(s.MPI_SOURCE, _FLIB_INITIAL_COMM));

	//int target_rank = data[2];

	uint32_t app_data_size = (*desc)->app_data_size;
	uint32_t epoch_data_size = (*desc)->epoch_data_size;
	int cp_data_unit = (*desc)->cp_data_disp_unit;

	uint32_t xor_data_offset = (*desc)->xor_data_offset;
	uint64_t* checkpoint_data = (uint64_t*)(*desc)->cp_data;
	uint64_t* xored_data = &((uint64_t*)(*desc)->cp_data)[xor_data_offset];

#ifdef FLIB_DEMAND_DEBUG
	debug_message_gen( FALSE, stderr, "*** CP: received demand PUT checkpoint done confirmation for CM [%i,IN] from CM [%i,IN]\n", target_rank, s.MPI_SOURCE);
#endif

	foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, s.MPI_SOURCE, 0, *(*desc)->cp_win);

	// TODO: for now, to make it simple, I copy in one bulk. It can be done in a smarter way using smaller blocks

	/* first, we copy the xored data from CM process (application data + epoch data) */
	foMPI_Get(checkpoint_data, (app_data_size + epoch_data_size) / cp_data_unit, MPI_UINT64_T, s.MPI_SOURCE,
			xor_data_offset, (app_data_size + epoch_data_size) / cp_data_unit, MPI_UINT64_T, *(*desc)->cp_win);
	foMPI_Win_unlock(s.MPI_SOURCE, *(*desc)->cp_win);	

	/* then, we calculate the xor of the data got from compute process and the xor data stored locally */
	calculate_xor(checkpoint_data, xored_data, app_data_size + epoch_data_size);

	/* we send the confirmation to the requesting process that the checkpoint is done and it can clear the logs */	

	assert(!gdsl_queue_is_empty((*cp_desc)->queued_demand_checkpoint_requests[s.MPI_SOURCE]));
	while(!gdsl_queue_is_empty((*cp_desc)->queued_demand_checkpoint_requests[s.MPI_SOURCE])) {
		int* proc_to_send = (int*) gdsl_queue_remove((*cp_desc)->queued_demand_checkpoint_requests[s.MPI_SOURCE]);
		MPI_Send(&(s.MPI_SOURCE), 1, MPI_INT, *proc_to_send, CP_DEMAND_PUT_CHECKPOINT_DONE, (*desc)->initial_user_comm);

#ifdef FLIB_DEMAND_DEBUG
		debug_message_gen( FALSE, stderr, "*** CP: sent demand PUT checkpoint done to %i (for %i)\n", *proc_to_send, s.MPI_SOURCE);
#endif

		free(proc_to_send);
	}
}

//#######################################################################################################################################
//GET demand checkpoint requests
//#######################################################################################################################################

void handle_cm_demand_get_checkpoint_request() {
	if(!do_logging()) {
		debug_message_gen_e(TRUE, stderr,"########### Cannot receive cm demand GET checkpoint request if no logging enables!\n");
	}
	MPI_Status s;
	int target_rank;

	MPI_Recv(&target_rank, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);

#ifndef	MEASURE_LOGGING_OVERHEADS
	assert(s.MPI_TAG == CM_DEMAND_GET_CHECKPOINT_REQUEST);
	assert(is_cm_proc(s.MPI_SOURCE));	
	assert(is_our_proc(target_rank, _FLIB_INITIAL_COMM));
#endif

#ifdef FLIB_DEMAND_DEBUG
	debug_message_gen( FALSE, stderr, "*** CP: received demand GET checkpoint request for CM [%i,IN] from CM [%i,IN]\n", target_rank, s.MPI_SOURCE);
#endif

	//this time we cannot queue requests, because each request has to be sent back with a (probably) unique epoch nr
	MPI_Send(&s.MPI_SOURCE, 1, MPI_INT, target_rank, CP_DEMAND_GET_CHECKPOINT_REQUEST, (*desc)->initial_user_comm);

#ifdef FLIB_DEMAND_DEBUG
	debug_message_gen( FALSE, stderr, "*** CP: sent demand GET checkpoint request to CM [%i,IN]\n", target_rank);
#endif
}

void handle_cm_demand_get_checkpoint_done() {
	if(!do_logging()) {
		debug_message_gen_e(TRUE, stderr,"########### Cannot receive cm demand GET checkpoint done if no logging enables!\n");
	}

	MPI_Status s;
	int data[2];
	int target_rank;
	int epoch_to_clear;

	//this time 'target_rank' means the process that originally requested a demand checkpoint
	MPI_Recv(&data[0], 2, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);
	target_rank = data[0];
	epoch_to_clear = data[1];

	assert(is_cm_proc(s.MPI_SOURCE));
	assert(s.MPI_TAG == CM_DEMAND_GET_CHECKPOINT_DONE);
	assert(is_our_proc(s.MPI_SOURCE, _FLIB_INITIAL_COMM));

	//TODO: the code below is identical to the code from put demand checkpoints; merge the two

	uint32_t app_data_size = (*desc)->app_data_size;
	uint32_t epoch_data_size = (*desc)->epoch_data_size;
	int cp_data_unit = (*desc)->cp_data_disp_unit;

	uint32_t xor_data_offset = (*desc)->xor_data_offset;
	uint64_t* checkpoint_data = (uint64_t*)(*desc)->cp_data;
	uint64_t* xored_data = &((uint64_t*)(*desc)->cp_data)[xor_data_offset];

#ifdef FLIB_DEMAND_DEBUG
	debug_message_gen( FALSE, stderr, "*** CP: received demand GET checkpoint done confirmation for CM [%i,IN] from CM [%i,IN]\n", target_rank, s.MPI_SOURCE);
#endif

	foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, s.MPI_SOURCE, 0, *(*desc)->cp_win);

	/* first, we copy the xored data from CM process (application data + epoch data) */
	foMPI_Get(checkpoint_data, (app_data_size + epoch_data_size) / cp_data_unit, MPI_UINT64_T, s.MPI_SOURCE,
			xor_data_offset, (app_data_size + epoch_data_size) / cp_data_unit, MPI_UINT64_T, *(*desc)->cp_win);
	foMPI_Win_unlock(s.MPI_SOURCE, *(*desc)->cp_win);	

	/* then, we calculate the xor of the data got from compute process and the xor data stored locally */
	calculate_xor(checkpoint_data, xored_data, app_data_size + epoch_data_size);

	/* we send the confirmation to the requesting process that the checkpoint is done and it can clear the logs */	
	data[0] = s.MPI_SOURCE;
	MPI_Send(&data[0], 2, MPI_INT, target_rank, CP_DEMAND_GET_CHECKPOINT_DONE, (*desc)->initial_user_comm);

#ifdef FLIB_DEMAND_DEBUG
	debug_message_gen( FALSE, stderr, "*** CP: sent demand GET checkpoint done to %i (for %i)\n", target_rank, s.MPI_SOURCE);
#endif
}

//#######################################################################################################################################
//NECESSARY checkpoint requests
//#######################################################################################################################################

void handle_cm_necessary_checkpoint_done() {
	if(!do_logging()) {
		debug_message_gen_e(TRUE, stderr,"########### ERROR: Cannot receive cm necessary checkpoint request if no logging enables!\n");
	}
	MPI_Status s;
	int data[2];

	MPI_Recv(&data[0], 2, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);

	assert(is_cm_proc(s.MPI_SOURCE));
	assert(s.MPI_TAG == CM_NECESSARY_CHECKPOINT_DONE);
	assert(is_our_proc(s.MPI_SOURCE, _FLIB_INITIAL_COMM));

	int target_rank = data[1];
	int world_target_rank = translate_from_flib_to_world_rank(target_rank);

	uint32_t app_data_size = (*desc)->app_data_size;
	uint32_t epoch_data_size = (*desc)->epoch_data_size;
	int cp_data_unit = (*desc)->cp_data_disp_unit;

	uint32_t xor_data_offset = (*desc)->xor_data_offset;
	uint64_t* checkpoint_data = (uint64_t*)(*desc)->cp_data;
	uint64_t* xored_data = &((uint64_t*)(*desc)->cp_data)[xor_data_offset];

#ifdef FLIB_DEMAND_DEBUG
	debug_message_gen( FALSE, stderr, "*** CP: received NECESSARY checkpoint done confirmation for CM [%i,IN] from CM [%i,IN]\n",
			world_target_rank, s.MPI_SOURCE);
#endif

	foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, s.MPI_SOURCE, 0, *(*desc)->cp_win);

	/* TODO: for now, to make it simple, I copy in one bulk. It can be done in a smarter way using smaller blocks
		TODO: I assume that a CP process does not have to lock its checkpoint and epoch-checkpoint window
		because there should not be any access from outside. 
		but - what about the recovery procedure??? */

	/* first, we copy the xored data from CM process (application data + epoch data) */
	foMPI_Get(checkpoint_data, (app_data_size + epoch_data_size) / cp_data_unit, MPI_UINT64_T, s.MPI_SOURCE,
			xor_data_offset, (app_data_size + epoch_data_size) / cp_data_unit, MPI_UINT64_T, *(*desc)->cp_win);
	foMPI_Win_unlock(s.MPI_SOURCE, *(*desc)->cp_win);	

	/* then, we calculate the xor of the data got from compute process and the xor data stored locally */
	calculate_xor(checkpoint_data, xored_data, app_data_size + epoch_data_size);

	/* we send the confirmation to the targeted process that the checkpoint is done and it can clear the logs */	

	//TODO: check it
	//data[2] = s.MPI_SOURCE;
	//MPI_Send(&data[0], 2, MPI_INT, world_target_rank, CP_NECESSARY_CHECKPOINT_DONE, (*desc)->initial_user_comm);

#ifdef FLIB_DEMAND_DEBUG
	debug_message_gen( FALSE, stderr, "*** CP: sent NECESSARY checkpoint done to %i (for %i)\n", world_target_rank, s.MPI_SOURCE);
#endif
}

//#######################################################################################################################################
//logs cleared messages
//#######################################################################################################################################

void handle_cm_global_logs_cleared() {
	if(!do_logging()) {
		debug_message_gen_e(TRUE, stderr,"########### ERROR: Cannot receive a confirmation about clearing global logs if no logging enables!\n");
	}
	MPI_Status s;
	int rank_to_notify;

	MPI_Recv(&rank_to_notify, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);

	assert(is_cm_proc(s.MPI_SOURCE));
	assert(s.MPI_TAG == CM_GLOBAL_LOGS_CLEARED);
	assert(is_our_proc(s.MPI_SOURCE, _FLIB_INITIAL_COMM));

	MPI_Send(&(s.MPI_SOURCE), 1, MPI_INT, rank_to_notify, CP_GLOBAL_LOGS_CLEARED, (*desc)->initial_user_comm);

#ifdef FLIB_DEMAND_DEBUG
	debug_message_gen( FALSE, stderr, "*** CP: sent confirmation about cleared global logs to %i (from %i)\n", rank_to_notify, s.MPI_SOURCE);
#endif
}

//#######################################################################################################################################
////////////////error and recovery requests
//#######################################################################################################################################

//#######################################################################################################################################
//TODO: A routine that reacts to a notification sent by a CM that realized that other CM is dead
//#######################################################################################################################################

void handle_proc_error() {
	MPI_Status s;
	int dead_rank;
	MPI_Recv(&dead_rank, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);

	if((*desc)->coordinated_checkpoint_started == TRUE) {
		debug_message_gen_e( TRUE, stderr, "############ a crash during a coordinated checkpoint. !!!! This message should not be displayed: check the code !!!!\n");
	}
}

//#######################################################################################################################################
//begins a recovery of a crashed process
//#######################################################################################################################################

void handle_proc_recovery() {
	MPI_Status s;
	int rank;
	MPI_Recv(NULL, 0, MPI_CHAR, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);

	assert(is_cm_proc(s.MPI_SOURCE));
	assert(s.MPI_TAG == CM_PROC_GLOBAL_RECOVERY || s.MPI_TAG == CM_PROC_LOCAL_RECOVERY);

	if((*desc)->coordinated_checkpoint_started == TRUE) {
		debug_message_gen_e( TRUE, stderr, "############ cannot handle a crash during the coordinated checkpoint; exiting...\n");
	}

	int process_being_recovered = s.MPI_SOURCE;

#ifdef FLIB_GLOBAL_RECOVERY_DEBUG
	debug_message_gen( FALSE, stderr, "^^^ received %s recovery request from %i\n", (s.MPI_TAG == CM_PROC_GLOBAL_RECOVERY) ? "GLOBAL" : "LOCAL", s.MPI_SOURCE);
#endif

	if(s.MPI_TAG == CM_PROC_GLOBAL_RECOVERY) {
		(*desc)->global_recovery_started = TRUE;

		send_cp_master_notification(CP_MASTER_GLOBAL_RECOVERY_REQUEST);
		send_cp_notification_to_assigned_cms(CP_GLOBAL_RECOVERY_REQUEST);
		wait_for_cm_global_recovery_confirmations();
#ifdef FLIB_GLOBAL_RECOVERY_DEBUG
		debug_message_gen_e( FALSE, stderr,"^^^ CP: recovery confirmations from CMs DONE, waiting for other CPs \n");
#endif
		fall_back_to_previous_checkpoint(process_being_recovered);

		finish_global_recovery();

		(*desc)->global_recovery_started = FALSE;
	}
}

//#######################################################################################################################################
//a routine called by other CPs that received a recovery request from the CP that started the recovery
//#######################################################################################################################################

void handle_cp_master_global_recovery_request() {
	assert(cp_proc());

	MPI_Status s;
	MPI_Recv(NULL, 0, MPI_CHAR, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);

	assert(is_cp_proc(s.MPI_SOURCE));
	assert(s.MPI_TAG == CP_MASTER_GLOBAL_RECOVERY_REQUEST);

	(*desc)->global_recovery_started = TRUE;

#ifdef FLIB_GLOBAL_RECOVERY_DEBUG	
	debug_message_gen_e( FALSE, stderr,"^^^ CP: received GLOBAL RECOVERY request from master \n");	
#endif

	send_cp_notification_to_assigned_cms(CP_GLOBAL_RECOVERY_REQUEST);
	wait_for_cm_global_recovery_confirmations();	//TODO: maybe unnecessary, make sure it's enough to lock and get

#ifdef FLIB_GLOBAL_RECOVERY_DEBUG
	debug_message_gen_e( FALSE, stderr,"^^^ CP: recovery confirmations from CMs DONE, waiting for other CPs \n");
#endif

	finish_global_recovery();

	(*desc)->global_recovery_started = FALSE;

#ifdef FLIB_GLOBAL_RECOVERY_DEBUG
	debug_message_gen_e( FALSE, stderr,"^^^ CP: recovery barrier passed! \n");
#endif
}

//#######################################################################################################################################
//In this routine CPs wait until all CMs get the recovery notification
//#######################################################################################################################################

void wait_for_cm_global_recovery_confirmations() {
	int procs_done = 0;
	int assigned_procs_nr = assigned_cm_procs_nr();
	int requests = FALSE;
	MPI_Status s;

	while(procs_done < assigned_procs_nr) {
		MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &requests, &s);
		if(requests == TRUE) {
			requests = FALSE;
			int tag = s.MPI_TAG;

			if(tag == CM_GLOBAL_RECOVERY_CONFIRMATION) {
				procs_done++;
				MPI_Recv(NULL, 0, MPI_CHAR, s.MPI_SOURCE, s.MPI_TAG, (*desc)->initial_user_comm, &s);
			}
			//skip most of other messages, we don't care about them at that point
			else if(tag == CM_COORDINATED_CHECKPOINT || tag == CM_FAKED_COORDINATED_CHECKPOINT) {
				uint64_t* faked_data = malloc((*desc)->app_data_size);
				MPI_Recv(faked_data, (*desc)->app_data_size / (*desc)->cp_data_disp_unit, MPI_UINT64_T, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, MPI_STATUS_IGNORE);
				free(faked_data);
				debug_message_gen_e( FALSE, stderr,"############ Skipping a coordinated checkpoint message \n");
			}
			else if(tag == CP_MASTER_COORDINATED_CHECKPOINT_REQUEST) {
				debug_message_gen_e( TRUE, stderr,"############ Sth went wrong, cannot perform a global recovery and do a global checkpoint at the same time \n");
			}
			else if(tag == CM_DEMAND_PUT_CHECKPOINT_REQUEST) {
				int faked_data;
				MPI_Recv(&faked_data, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, MPI_STATUS_IGNORE);
				debug_message_gen_e( FALSE, stderr,"############ Skipping a demand checkpoint request \n");
			}
			else if(tag == CM_DEMAND_PUT_CHECKPOINT_DONE || tag == CM_NECESSARY_CHECKPOINT_DONE) {
				int faked_data[3];
				MPI_Recv(&faked_data[0], 3, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, MPI_STATUS_IGNORE);
				debug_message_gen_e( FALSE, stderr,"############ Skipping a demand checkpoint done\n");
			}
			else if(tag == CM_PROC_ERROR) {//TODO: address this. For example, more than one proc may notice than a single process is dead
				int dead_rank;
				MPI_Recv(&dead_rank, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);
				debug_message_gen_e( TRUE, stderr,"############ TODO \n");
			}
			else if(tag == CM_PROC_GLOBAL_RECOVERY) {
				debug_message_gen_e( TRUE, stderr,"############ Cannot handle more than two crashes in one CP group, exiting... \n");
			}
			else if(tag == CP_MASTER_GLOBAL_RECOVERY_REQUEST) {
				debug_message_gen_e( TRUE, stderr,"############ Sth went wrong, a global recovery is already in progress, exiting... \n");
			}
			else if(tag == CM_QUIT) {
				debug_message_gen_e( TRUE, stderr,"############ CM_QUIT during global recovery, TODO...\n");
			}
			else {
				handle_unknown_request_cp(tag);
			}
		}
	}
}

//#######################################################################################################################################
//A scheme called to finish the recovery. It restores the state of the failed process from checksums
//#######################################################################################################################################

void fall_back_to_previous_checkpoint(int process_being_recovered) {
	int app_data_size = (*desc)->app_data_size;
	int cp_data_unit = (*desc)->cp_data_disp_unit;
	int* assigned_procs = assigned_cm_procs();

	//TODO: check if we can do it in a more memory-efficient way
	uint64_t* remote_data = (uint64_t*) malloc( app_data_size * sizeof(char) );
	uint64_t* xor_result = (uint64_t*) malloc( app_data_size * sizeof(char) );

	uint32_t xor_data_offset = (*desc)->xor_data_offset;
	uint64_t* xored_data = &((uint64_t*)(*desc)->cp_data)[xor_data_offset];

	memset(remote_data, 0, app_data_size);	
	memcpy(xor_result, xored_data, app_data_size);

	int j = 0;
	for(; j < assigned_cm_procs_nr(); j++) {
		if(assigned_procs[j] == process_being_recovered) {
			continue;
		}

		foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, assigned_procs[j], 0, *(*desc)->cp_win);

		foMPI_Get(remote_data, app_data_size / cp_data_unit, MPI_UINT64_T, assigned_procs[j], 0, app_data_size / cp_data_unit, MPI_UINT64_T, *(*desc)->cp_win);
		foMPI_Win_flush(assigned_procs[j], *(*desc)->cp_win);

		foMPI_Win_unlock(assigned_procs[j], *(*desc)->cp_win);

		calculate_xor(remote_data, xor_result, app_data_size);
	}

	MPI_Send(xor_result, app_data_size / cp_data_unit, MPI_UINT64_T, process_being_recovered, CP_RECOVERY_CHECKPOINT, (*desc)->initial_user_comm);

	free(assigned_procs);
	free(xor_result);
	free(remote_data);
}

//#######################################################################################################################################
//A scheme that handles a CM quitting
//#######################################################################################################################################

void handle_quit(int* procs_nr) {
	MPI_Status s;
	MPI_Recv(NULL, 0, MPI_CHAR, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);

	assert(is_cm_proc(s.MPI_SOURCE));
	assert(is_our_proc(s.MPI_SOURCE, _FLIB_INITIAL_COMM ) == TRUE);

	(*procs_nr)++;

#ifdef FLIB_VERBOSE_DEBUG
	if(*procs_nr == (*desc)->group_size - 1) {
		debug_message_gen( FALSE, stderr,"QQQ CP: received quit notification from %i (%i procs from my group quit)\n", s.MPI_SOURCE, *procs_nr);
	}
#endif
}

int handle_unknown_request_cp(int tag) {
	debug_message_gen( TRUE, stderr,"############ something went wrong, I received a message with a tag: %i\n", tag);
}
