#ifndef _FLIB_INTERNAL_
#define _FLIB_INTERNAL_

#include <mpi.h>
#include <string.h>

#include "nbc.h"
#include "fompi.h"
#include "gdsl.h"

#include <unistd.h>
#include <stdarg.h>

#define _FLIB_NONE_HIERARCHY 	888
#define _FLIB_NODES_HIERARCHY	889

#define _FLIB_FIRST_IN_INITIAL 	11111
#define _FLIB_FIRST_IN_FLIB_CM	11112
#define _FLIB_FIRST_IN_FLIB_CP	11113
#define _FLIB_FIRST_IN_NODE		 	11114
#define _FLIB_ANY_IN_INITIAL 		11115
#define _FLIB_ANY_IN_FLIB_CM    11116
#define _FLIB_ANY_IN_FLIB_CP    11117
#define _FLIB_ANY_IN_NODE    		11118

#define _FLIB_INITIAL_COMM			444
#define _FLIB_FLIB_COMM					445
#define _FLIB_NODE_COMM					446

#define STATE_GLOBAL_RECOVERY 				8989
#define STATE_COORDINATED_CHECKPOINT	8787
#define STATE_NORMAL									8686

#define _FLIB_DEMAND_PUT_NOT_SENT			-10

/* used in logging */
/* we log only remote memory access operations */
typedef enum Operation_type_t {
	PUT,
	ACCUMULATE,
	GET,
	GET_ACCUMULATE,
	CAS,
	FETCH_AND_OP

} Operation_type;

//#######################################################################################################################################
//structures that are logged puts/gets
//#######################################################################################################################################

/* the structure containing the common parameters of all the operations  */
typedef struct Common_params_t {
	int epoch_nr;
	int target_rank;	

	//the size of the whole log
	int next_node_offset;
	Operation_type op_type;

	//respective counters
	int GNC;
	int GC;
	int SC;

	//if this is a combing or a replacing put
	int combining;

} Common_params;

/* the structure containing the arguments of put/get operations */
typedef struct Put_get_args_t {
	int origin_count;
	MPI_Datatype origin_datatype;
	MPI_Aint target_disp;
	int target_count;	
	MPI_Datatype target_datatype;

} Put_get_args;

/* the structure containing the arguments of accumulate operations */
typedef struct Acc_args_t {
	int origin_count;
	MPI_Datatype origin_datatype;
	MPI_Aint target_disp;
	int target_count;	
	MPI_Datatype target_datatype;
	MPI_Op operation;

} Acc_args;

/* the structure for storing get arguments before actual remote logging */
typedef struct Local_get_info_t {
	const void* origin_addr;
	int origin_count;
  MPI_Datatype origin_datatype;
	int target_rank;
  MPI_Aint target_disp;
  int target_count;
  MPI_Datatype target_datatype;
} Local_get_info;

//#######################################################################################################################################
//The logging space
//#######################################################################################################################################

/* the structure encompassing the whole logging space of every process */
typedef struct Log_space_t {
	foMPI_Win* local_list;
	foMPI_Win* global_list;

	//local logs of puts
	char* local_list_data;

	//local logs of gets
	char* global_list_data;

	/* the window for storing the control data of global logs*/
	/* these are: 
			- global sums
			- next free offset
			- number of nodes */
	foMPI_Win* global_logs_control_win;

	/* the control data of the global logs bound to the window: */
	/* 0..|CM|-1  - the global sums */
	/* |CM|			  - next free global offset */
	/* |CM| + 1   - number of nodes global */
	int64_t* global_logs_control_data;

	/* an array where a process records which processes it communicated with using gets */
	//TODO: double check if necessary and if we can remove them
	int64_t* procs_communicated_using_gets;
	int64_t* procs_communicated_using_puts;

	//TODO: double check if necessary
	foMPI_Win* puts_win;

	/* array of queues containing gets that have to be logged during a synchronization operation*/
	gdsl_queue_t* gets_to_be_logged;

	//different fields for the local logging space of puts
	/* local logs are not modified by remote processes, thus we don't need a separate control window */	
	int64_t* local_sums;
	int next_free_local_offset;
	int number_of_nodes_local;

	//TODO: reduce the nr of windows
	foMPI_Win* local_list_control;
	foMPI_Win* local_list_size;

	//fields that control the testing logs
	int __cleared_global_logs;
	int __cleared_global_test_logs;
}	Log_space;

//#######################################################################################################################################
//various fLIB parameters used in logging
//#######################################################################################################################################

typedef struct fLIB_Parameters_t {
	double puts_relative_demand_threshold;
	double gets_relative_demand_threshold;
	int log_puts;
	int log_gets;
	int log_list_size;
} fLIB_Parameters;

//#######################################################################################################################################
//main descriptor
//#######################################################################################################################################

/* A general descriptor data structure used by both CM and CP processes to encapsulate some important information */
typedef struct fMPI_Descriptor {
	/* fields for quick access to the properties of the initial user communicator */
	int initial_user_rank;
	int initial_user_size;
	MPI_Comm initial_user_comm;

	fLIB_Parameters* flib_parameters;

	int state;

	/* the number of CM and CP procs */
	int cm_size;
	int cp_size;

	/* the rank inside the flib communicator  */
	/* TODO: used only by CM procs (for now). Maybe move it to a CM-specific descriptor? */
	int flib_level_rank;

	/* the flib communicator */
  /* TODO: used only by CM procs (for now). Move it to a CM-specific descriptor? */
	MPI_Comm* flib_level_comm;

	int node_level_rank;
	int node_level_size;
	MPI_Comm node_level_comm;
	MPI_Group node_level_group;

	/* useful for CP procs which do not have direct access to the application and epoch window */
	/* they use it while doing the checkpointing */
	int app_data_size; 
	int epoch_data_size;
	int counters_size;

	/* checkpoint window and checkpoint data */
	/* there is one window for both application data and epoch counters */
	foMPI_Win* cp_win;
	void* cp_data;

	/* different offsets pointing to different parts of the checkpoint data in the checkpoint window */
	uint32_t data_offset;
	uint32_t app_data_offset;
	uint32_t epoch_data_offset;
	uint32_t counters_offset;

	uint32_t xor_data_offset;	
	uint32_t xor_app_data_offset;
	uint32_t xor_epoch_data_offset;
	uint32_t xor_counters_offset;

	/* displacement unit in cp window (stored here for easy access) */
	int cp_data_disp_unit;	

	/* fields for quick access of the group information */
	int groups_nr;
	int group_size;
	
	/* a field determining whether the current process is a computing process or a CP process */
	int compute_proc;

	/* are there any awaiting demand checkpoint requests? */
	int awaiting_requests;

	/* timing parameters */
	double initial_time;
	double prev_cp_time;
	double daly_interval;

	/* type of coordinated protocol */
	int coordinated_protocol_type;

	//fields that determine if we started a CC or a recovery
	int coordinated_checkpoint_started;
	int global_recovery_started;

	//whether we degenerated to CC only
	int degenerated;

	//number of remote memory accesses performed
	int remote_ops_done;

	//t-awareness policy
	int hierarchy_scheme;

	//how do we pick processes to perform a demand checkpoint
	int demand_target_selection_heuristics;

	//a protocol for logging gets remotely (currently only locking)
	int logging_gets_scheme;

	//TODO: now not used, to be implemented
	//determines if demand checkpoints are separate or merged with CCs
	int demand_checkpointing_merge_requests;

	//the type of the user application
	//treated as a hint that helps to optimize some parts of our protocol
	int app_type;
} fMPI_Desc_t;

typedef fMPI_Desc_t* fMPI_Desc;

//#######################################################################################################################################
//the CP descriptor
//#######################################################################################################################################

/* CP Descriptor - holds information specific to CP processes */

typedef struct fMPI_CP_Descriptor {
	//used while doing CC to check if CC
	//do not begin to overlap
	int received_coordinated_checkpoints_nr;
	int* received_coordinated_checkpoint;

	//used to queue put demand checkpoint requests
	//to avoid wasting bandwidth
	int** sent_demand_checkpoint_requests;
	gdsl_queue_t* queued_demand_checkpoint_requests;
	
} fMPI_CP_Desc_t;

typedef fMPI_CP_Desc_t* fMPI_CP_Desc;

//#######################################################################################################################################
//the CM descriptor
//#######################################################################################################################################

/* CM Descriptor - holds information specific to CM processes */

typedef struct fMPI_CM_Descriptor {

	/* a pointer to the user application window */
	foMPI_Win* app_win;

	/* a pointer to the user application data*/
	void* app_ptr;

	/* epoch window - the global counter*/
	//TODO: double check if can be removed
	foMPI_Win* epoch_win;

	/* epoch data connected with the epoch window. */
	/* this is the global counter updated by remote processes */
	//TODO: double check if can be removed
	int64_t* epoch_data;
	
	/* this is the local counter updated only by one process */
	/* it is recorded during checkpoints, but it is not connected to any window */
	int64_t* local_epoch_counter;

	//other counters
	int64_t GNC_cnt;
	int64_t SC_cnt;
	int64_t GC_cnt;

	/* logging space */
	Log_space* log_space;

	int* sent_checkpoint_request;
	int* sent_checkpoint_request_epoch;

	int* sent_necessary_checkpoint_done;

	/* request used while synchronizing for taking coordinated checkpoint */
	NBC_Handle coordinated_checkpoint_request;

	/* request used while synchronizing for global recovery */
	NBC_Handle global_recovery_request;

	//nr of sent put demand checkpoints
	//TODO: change the name to a better one
	int sent_cm_adaptive_requests;

	//nr of necessary checkpoints sent
	int sent_cm_necessary_requests;

	//nr of get demand checkpoints sent
	int sent_cm_get_demand_requests;

	//a current superstep nr
	int superstep;

	//the frequency of CC (per how many fences we take CC)
	int checkpoint_frequency;

	//a field that says if we're in the period of measuring how long it takes to perform a CC
	int initial_coordinated_measurement;

	//whether the user wants us to recompute the frequency of CC at some point
	int recompute_coordinated_checkpoint_frequency;

	//fields that are used in deciding whether we start to recompute the CC interval
	double avg_interval;
	int nr_of_intervals_for_avg;
	int nr_of_irregular_intervals;
	int initial_intervals_for_skipping;
	double frequency_interval_threshold;
	int nr_interval_threshold;

	//the M marker data structures
	foMPI_Win* marker_M_win;
	int* marker_M_space;
} fMPI_CM_Desc_t;

typedef fMPI_CM_Desc_t* fMPI_CM_Desc;

extern fMPI_Desc* desc;
extern fMPI_CM_Desc* cm_desc;
extern fMPI_CP_Desc* cp_desc;

#endif

