#include "flib_win_create.h"
#include "flib_cp_proc.h"
#include "flib_debug.h"
#include <math.h>

/* initialization of our global descriptors
	 it would probably be more elegant to pass it as an argument
	 but if it's global, the protocols can be more transparent to the user */

fMPI_Desc* desc = NULL;
fMPI_CM_Desc* cm_desc = NULL;
fMPI_CP_Desc* cp_desc = NULL;

//#######################################################################################################################################
//general routines for window allocation/creation
//#######################################################################################################################################

int fMPI_Win_allocate(MPI_Aint size, int disp_unit, MPI_Info info, MPI_Comm comm, void *baseptr, foMPI_Win *win, int cp_procs, MPI_Comm* user_comm, int hierarchy, double mtbf, int protocol_type, int log_puts, int log_gets, int log_list_size) {

	return initialize_window("win_allocate", baseptr, size, disp_unit, info, comm, win, cp_procs, user_comm, hierarchy, mtbf, protocol_type, log_puts, log_gets, log_list_size);
} 

int fMPI_Win_create(void *base, MPI_Aint size, int disp_unit, MPI_Info info, MPI_Comm comm, foMPI_Win *win, int cp_procs, MPI_Comm* user_comm, int hierarchy, double mtbf, int protocol_type, int log_puts, int log_gets, int log_list_size) {

	return initialize_window("win_create", base, size, disp_unit, info, comm, win, cp_procs, user_comm, hierarchy, mtbf, protocol_type, log_puts, log_gets, log_list_size);
}

int initialize_window(const char* creation_type, void *base, MPI_Aint size, int disp_unit, MPI_Info info, MPI_Comm comm, foMPI_Win *win, int cp_procs,
		MPI_Comm* user_comm, int hierarchy, double mtbf, int protocol_type, int log_puts, int log_gets, int log_list_size) {

	int result = MPI_SUCCESS;

	// allocate the memory for the flib descriptor and initialize some general fields
	init_flib_desc(hierarchy, protocol_type, log_puts, log_gets, log_list_size);

	divide_into_groups(hierarchy, cp_procs, comm, user_comm);
	debug_message_e( _FLIB_FIRST_IN_INITIAL, FALSE, stderr, ">>> Desc initialized & Groups divition done\n");

#ifdef FLIB_DEBUG
	__test_groups_division();
	debug_message_e( _FLIB_FIRST_IN_INITIAL, FALSE, stderr, ">>> Groups division test done\n");
#endif

	/* allocate the memory for descriptors specific to CP and CM processes */
	if(cm_proc()) {
		init_cm_descriptor();
		debug_message_e( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, ">>> CM Desc initialized\n");

		if((*desc)->flib_level_rank == 0) {
			print_comm_size((*desc)->flib_level_comm);
		}
	}
	else {
		init_cp_descriptor();
		debug_message_e( _FLIB_FIRST_IN_FLIB_CP, FALSE, stderr, ">>> CP Desc initialized\n");

		if((*desc)->flib_level_rank == 0) {
			print_comm_size((*desc)->flib_level_comm);
		}
	}

	init_checkpoint_window(size);
	debug_message_e( _FLIB_FIRST_IN_INITIAL, FALSE, stderr, ">>> Checkpoint window initialized\n");

	if(cm_proc()) {		
		if(!strcmp(creation_type, "win_allocate")) {
			result = foMPI_Win_allocate(size, disp_unit, info, *user_comm, base, win);
			debug_message_e( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, ">>> User window ALLOCATED\n");
		}
		else if(!strcmp(creation_type, "win_create")) {
			result = foMPI_Win_create(base, size, disp_unit, info, *user_comm, win);
			debug_message_e( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, ">>> User window CREATED\n");
		}
		else {
			debug_message_e( _FLIB_FIRST_IN_FLIB_CM, TRUE, stderr, "### unknown type of window intialization\n");
		}
		debug_message_e( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, ">>> User window initialized\n");

		init_cm_windows(win);
		debug_message_e( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, ">>> CM-specific windows initialized\n");
	}

	init_timing(mtbf);
	debug_message_e( _FLIB_FIRST_IN_INITIAL, FALSE, stderr, ">>> Timing initialized\n");

	initiate_coordinated_checkpointing();
	debug_message_e( _FLIB_FIRST_IN_INITIAL, FALSE, stderr, ">>> Coordinated checkpointing initialized\n");

	initiate_global_recovery_protocol();
	debug_message_e( _FLIB_FIRST_IN_INITIAL, FALSE, stderr, ">>> Global recovery protocol initialized\n");

	take_initial_checkpoint();
	debug_message_e( _FLIB_FIRST_IN_INITIAL, FALSE, stderr, ">>> Initial checkpoint taken\n");

	if((*desc)->initial_user_rank == 0) {
		print_flib_parameters();
		debug_message_e( _FLIB_FIRST_IN_INITIAL, FALSE, stderr, ">>> Parameters printed\n");
	}

	start_timing();
	debug_message_e( _FLIB_FIRST_IN_INITIAL, FALSE, stderr, ">>> Timing started\n");

	if(cm_proc()) {
#if defined(FLIB_LOGGING_PUTS_DEBUG) || defined(FLIB_LOGGING_GETS_DEBUG)
		if(do_logging()) {
			__init_logging_testing();
			debug_message_e( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, ">>> Testing logging initialized\n");
		}
#endif

		debug_message_e( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, ">>> CM Initializtion complete\n");
		return result;
	}
	else {
		debug_message_e( _FLIB_FIRST_IN_FLIB_CP, FALSE, stderr, ">>> CP Initializtion complete\n");
		cp_process_notifications();

		return result;
	}
}

//#######################################################################################################################################
//Timing and Daly-related routines
//#######################################################################################################################################

void init_timing(double mtbf) {
	if((*desc)->coordinated_protocol_type == FLIB_NO_COORDINATED_SCHEME || (*desc)->coordinated_protocol_type == FLIB_ORDERED_CHECKPOINT) {
		(*desc)->daly_interval = 0;
		return;
	}

	if(mtbf < 0) {
		//mtbf < 0 means that the user explicitly passes its chosen interval equal to abs(mtbf)
		(*desc)->daly_interval = -1*mtbf;
	}
	else if(mtbf == 0) {
		debug_message_gen_e( TRUE, stderr, "### MTBF cannot be equal 0\n");
	}
	else {
		//any mtbf > 0 means that the user wants to calculate the daly interval
		calculate_daly_interval(mtbf);
	}

	assert((*desc)->daly_interval > 0);
}

void start_timing() {
	//we have to start the timing right before the processes begin to process
	//different CM & CP sections of code
	MPI_Barrier((*desc)->initial_user_comm);

	//initially both prev and initial values are the same
	(*desc)->prev_cp_time = MPI_Wtime();
	(*desc)->initial_time = (*desc)->prev_cp_time;
}

void calculate_daly_interval(double mtbf) {
	int faked_checkpoints = 10;	//arbitrary value, TODO: parameter
	double* times = malloc( faked_checkpoints * sizeof(double));
	double time_stamp = -1;
	int i = 0;

	for(; i < faked_checkpoints; i++) {
		MPI_Barrier((*desc)->initial_user_comm);

		time_stamp = MPI_Wtime();
		take_daly_measurement_checkpoint();
		times[i] = MPI_Wtime() - time_stamp;

		debug_message( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, ">>>>>> %i faked checkpoint: time to take: %f\n", i, times[i]);
	}

	if((*desc)->initial_user_rank == 0) {
		double sum = 0;
		int warmups = 2;	//arbitrary value
		for(i = warmups; i < faked_checkpoints; i++) {
			sum += times[i];
		}

		//currently we just take an average
		//TODO: maybe improve by using more advanced stats
		double delta = sum / (double)(faked_checkpoints - warmups);

		if(delta < 2*mtbf) {
			(*desc)->daly_interval = sqrt(2*delta*mtbf) * (1 + (1/3)*sqrt(delta/(2*mtbf)) +(1/9)*(delta/(2*mtbf))) - delta;
		}
		else {
			(*desc)->daly_interval = mtbf;
		}
		free(times);
	}

	MPI_Bcast(&(*desc)->daly_interval, 1, MPI_DOUBLE, 0, (*desc)->initial_user_comm);
}

void take_daly_measurement_checkpoint() {
	if(cm_proc()) {
		int daly_measurement = TRUE;
		finalize_coordinated_checkpoint(daly_measurement);
	}
	else {
		MPI_Status s;
		int cm_procs_done = 0;
		int request = FALSE;

		while(cm_procs_done < assigned_cm_procs_nr() ) {
			MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &request, &s);

			if(request == TRUE) {
				request = FALSE;
				int tag = s.MPI_TAG;
				assert(tag == CM_FAKED_COORDINATED_CHECKPOINT);

				int daly_measurement = TRUE;
				handle_simple_bsp_coordinated_checkpoint(daly_measurement);
				cm_procs_done++;	
			}
		}
	}

	MPI_Barrier((*desc)->initial_user_comm);
}

//#######################################################################################################################################
//routines for initializing coordinated checkpointing
//#######################################################################################################################################

void initiate_coordinated_checkpointing() {
	if(cp_proc()) {
		(*cp_desc)->received_coordinated_checkpoints_nr = 0;

		int i = 0;
		for(; i < (*desc)->initial_user_size; i++) {
			((*cp_desc)->received_coordinated_checkpoint)[i] = FALSE;
		}
	}
	else {
		//if we use a naive BSP_DISTRIBUTED (which uses allreduce), we have to
		//call it once at the beginning because in libNBC the first non-blocking Ibarrier
		//is actually a blocking one
		if((*desc)->coordinated_protocol_type == FLIB_BSP_DISTRIBUTED) {
			MPI_Status s;

			NBC_Ibarrier(*(*desc)->flib_level_comm, &((*cm_desc)->coordinated_checkpoint_request) );
			NBC_Wait( &((*cm_desc)->coordinated_checkpoint_request), &s);
		}
	}
}

//#######################################################################################################################################
//routines for initializing recovery protocols
//#######################################################################################################################################

void initiate_global_recovery_protocol() {
	if(cm_proc()) {
		MPI_Status s;

		NBC_Ibarrier(*(*desc)->flib_level_comm, &((*cm_desc)->global_recovery_request) );
		NBC_Wait( &((*cm_desc)->global_recovery_request), &s);
	}
}

//#######################################################################################################################################
//printing all parameters
//#######################################################################################################################################

void print_flib_parameters() {
	debug_message_gen( FALSE, stderr, ">>>>>> CP Processes: %i\n", (*desc)->cp_size);

	if((*desc)->coordinated_protocol_type == FLIB_BSP_DISTRIBUTED) {
		debug_message_gen_e( FALSE, stderr, ">>>>>> Coordinated protocol type: BLOCKING ALL REDUCE scheme\n");
	}
	if((*desc)->coordinated_protocol_type == FLIB_BSP_SIMPLE) {
		debug_message_gen_e( FALSE, stderr, ">>>>>> Coordinated protocol type: BSP NAIVE FREQUENCY scheme\n");
	}
	else if((*desc)->coordinated_protocol_type == FLIB_NO_BSP) {
		debug_message_gen_e( FALSE, stderr, ">>>>>> Coordinated protocol type: NO BSP scheme\n");
	}
	else if((*desc)->coordinated_protocol_type == FLIB_NO_COORDINATED_SCHEME) {
		debug_message_gen_e( FALSE, stderr, ">>>>>> Coordinated protocol type: no coordinated scheme\n");
	}
	else if((*desc)->coordinated_protocol_type == FLIB_ORDERED_CHECKPOINT) {
		debug_message_gen_e( FALSE, stderr, ">>>>>> Coordinated protocol type: ordered scheme\n");
	}
	else {
		debug_message_gen_e( TRUE, stderr, "###### Unknown coordinated protocol type!\n");
	}

	debug_message_gen( FALSE, stderr, ">>>>>> Daly interval: %f\n", (*desc)->daly_interval);

	if((*desc)->hierarchy_scheme == _FLIB_NONE_HIERARCHY) {
		debug_message_gen_e( FALSE, stderr, ">>>>>> Hierarchy: NONE\n");
	}
	else if((*desc)->hierarchy_scheme == _FLIB_NODES_HIERARCHY) {
		debug_message_gen_e( FALSE, stderr, ">>>>>> Hierarchy: NODES\n");
	}
	else {
		debug_message_gen_e( TRUE, stderr, "###### Unknown hierarchy scheme!\n");
	}

	if(do_logging()) {
		debug_message_gen_e( FALSE, stderr, ">>>>>> Logging: YES\n");
		if((*desc)->demand_target_selection_heuristics == FLIB_AVERAGE) {
			debug_message_gen_e( FALSE, stderr, ">>>>>> Demand target heuristics selection: AVERAGE\n");
		}
		else if((*desc)->demand_target_selection_heuristics == FLIB_MAX) {
			debug_message_gen_e( FALSE, stderr, ">>>>>> Demand target heuristics selection: MAX\n");
		}
		else if((*desc)->demand_target_selection_heuristics == FLIB_ALL) {
			debug_message_gen_e( FALSE, stderr, ">>>>>> Demand target heuristics selection: ALL\n");
		}
		else {
			debug_message_gen_e( TRUE, stderr, "###### Unknown demand target selection heuristics scheme!\n");
		}

		debug_message_gen( FALSE, stderr, ">>>>>> Local/Remote logging list size [bytes]: %zu\n", log_list_size());

		if((*desc)->logging_gets_scheme == FLIB_LOGGING_GETS_LOCK) {
			debug_message_gen_e( FALSE, stderr, ">>>>>> Logging gets scheme: LOCK\n");
		}
		else if((*desc)->logging_gets_scheme == FLIB_LOGGING_GETS_ATOMICS) {
			debug_message_gen_e( FALSE, stderr, ">>>>>> Logging gets scheme: ATOMICS\n");
		}

		if((*desc)->flib_parameters->log_puts == TRUE) {
			debug_message_gen_e( FALSE, stderr, ">>>>>> Logging puts: YES\n");
		}
		else {
			debug_message_gen_e( FALSE, stderr, ">>>>>> Logging puts: NO\n");
		}

		if((*desc)->flib_parameters->log_gets == TRUE) {
			debug_message_gen_e( FALSE, stderr, ">>>>>> Logging gets: YES\n");
		}
		else {
			debug_message_gen_e( FALSE, stderr, ">>>>>> Logging gets: NO\n");
		}
	}
	else {
		debug_message_gen_e( FALSE, stderr, ">>>>>> Logging: NO\n");
	}
}

//#######################################################################################################################################
//routines for initializing main, CM, and CP descriptors
//#######################################################################################################################################

void init_flib_desc(int hierarchy, int protocol_type, int log_puts, int log_gets, int log_list_size) {
	desc = malloc( sizeof(fMPI_Desc) );
	assert(desc != NULL);

	*desc = malloc( sizeof(fMPI_Desc_t) );
	assert(*desc != NULL);

	(*desc)->flib_parameters = malloc( sizeof(fLIB_Parameters) );
	assert((*desc)->flib_parameters != NULL);

	(*desc)->flib_parameters->puts_relative_demand_threshold = 0.25;
	(*desc)->flib_parameters->gets_relative_demand_threshold = 0.5;
	(*desc)->flib_parameters->log_puts = log_puts;
	(*desc)->flib_parameters->log_gets = log_gets;
	(*desc)->flib_parameters->log_list_size = log_list_size;

	//state of a process: normal or being recovered!
	(*desc)->state = STATE_NORMAL;

	//awaiting demand checkpoint requests
	(*desc)->awaiting_requests = FALSE;

	//type of the protocol, or if none to be used
	(*desc)->coordinated_protocol_type = protocol_type;

	//we begin with a non-degenerated coordinated protocol
	(*desc)->degenerated = FALSE;

	//used while deciding if we want to degenerate
	(*desc)->remote_ops_done = 0;

	//what t-awareness to we use
	(*desc)->hierarchy_scheme = hierarchy;

	(*desc)->coordinated_checkpoint_started = FALSE;
	(*desc)->global_recovery_started = FALSE;

	//TODO: a parameter
	//what policy we use while deciding whom to select for a deman checkpoint target
	(*desc)->demand_target_selection_heuristics = FLIB_MAX;//FLIB_AVERAGE;

	//a protocol for remote logging of gets
	(*desc)->logging_gets_scheme = FLIB_LOGGING_GETS_LOCK;
	//TODO: this decides if we use separate checkpoints for CC and Demand or do we merge them
	//(*desc)->demand_checkpointing_merge = TRUE;

	//TODO: pass in as the application type
	//used as a hint while deciding if the user should
	//increase the size of the logging space
	(*desc)->app_type = FLIB_APP_BLK;
}

void init_cm_descriptor() {
	cm_desc = malloc( sizeof(fMPI_CM_Desc) );
	assert(cm_desc != NULL);

	*cm_desc = malloc( sizeof(fMPI_CM_Desc_t) );
	assert(*cm_desc != NULL);

	(*cm_desc)->sent_cm_adaptive_requests = 0;
	(*cm_desc)->sent_cm_necessary_requests = 0;
	(*cm_desc)->sent_cm_get_demand_requests = 0;
	(*cm_desc)->sent_checkpoint_request = malloc( sizeof(int) * (*desc)->initial_user_size );
	(*cm_desc)->sent_checkpoint_request_epoch = malloc( sizeof(int) * (*desc)->initial_user_size );

	assert((*cm_desc)->sent_checkpoint_request != NULL);
	assert((*cm_desc)->sent_checkpoint_request_epoch != NULL);

	(*cm_desc)->sent_necessary_checkpoint_done = malloc( sizeof(int) * (*desc)->initial_user_size );

	assert((*cm_desc)->sent_necessary_checkpoint_done != NULL);

	int i = 0;
	for(; i < (*desc)->initial_user_size; i++) {
		(*cm_desc)->sent_checkpoint_request[i] = FALSE;
		(*cm_desc)->sent_necessary_checkpoint_done[i] = FALSE;
		(*cm_desc)->sent_checkpoint_request_epoch[i] = _FLIB_DEMAND_PUT_NOT_SENT;
	}

	(*cm_desc)->checkpoint_frequency = -1;
	(*cm_desc)->superstep = 0;

	if((*desc)->coordinated_protocol_type == FLIB_NO_BSP || (*desc)->coordinated_protocol_type == FLIB_NO_COORDINATED_SCHEME) {
		(*cm_desc)->initial_coordinated_measurement = FALSE;
	}
	else {
		(*cm_desc)->initial_coordinated_measurement = TRUE;
	}

	//TODO: test & debug this adaptive scheme!
	(*cm_desc)->recompute_coordinated_checkpoint_frequency = FALSE;
	(*cm_desc)->nr_of_irregular_intervals = 0;
	(*cm_desc)->nr_of_intervals_for_avg = 0;
	(*cm_desc)->avg_interval = 0;
	(*cm_desc)->nr_interval_threshold = 4;

	/*(*cm_desc)->number_of_saved_previous_intervals = 10;
	if((*desc)->coordinated_protocol_type == FLIB_BSP_SIMPLE) {
		(*cm_desc)->simple_bsp_checkpoints_previous_intervals = malloc(sizeof(int) * (*cm_desc)->number_of_saved_previous_intervals);
	}*/

	//structures used if caching of demand checkpoint requests is enabled
	/*if((*desc)->demand_checkpointing_merge_requests == TRUE) {
		(*cm_desc)->cp_procs_counts = calloc((*desc)->cp_size, sizeof(int));
		assert((*cm_desc)->cp_procs_counts != NULL);
		(*cm_desc)->cp_procs_ranks_to_send = malloc(sizeof(gdsl_queue_t*) * (*desc)->cp_size);
		assert((*cm_desc)->cp_procs_ranks_to_send != NULL);
		
		int i = 0;
		for(; i < (*desc)->cp_size; i++) {
			((*cm_desc)->cp_procs_ranks_to_send)[i] = malloc(sizeof(gdsl_queue_t));
			assert(((*cm_desc)->cp_procs_ranks_to_send)[i] != NULL);
			((*cm_desc)->cp_procs_ranks_to_send)[i]->gdsl_queue_alloc(NULL,NULL,NULL);
		}
	}
	else {
		(*cm_desc)->cp_procs_counts = NULL;
		(*cm_desc)->cp_procs_ranks_to_send = NULL;
	}*/
}

void init_cp_descriptor() {
	cp_desc = malloc( sizeof(fMPI_CP_Desc) );
	assert(cp_desc != NULL);

	*cp_desc = malloc( sizeof(fMPI_CP_Desc_t) );
	assert(*cp_desc != NULL);

	(*cp_desc)->received_coordinated_checkpoints_nr = 0;

	// TODO: use smaller size!!! 
	(*cp_desc)->received_coordinated_checkpoint = malloc( sizeof(int) * (*desc)->initial_user_size);
	assert((*cp_desc)->received_coordinated_checkpoint != NULL);

	if(do_logging()) {
		//init demand requests queues
		//TODO: change the size
		(*cp_desc)->queued_demand_checkpoint_requests = malloc((*desc)->initial_user_size * sizeof(gdsl_queue_t));
		assert((*cp_desc)->queued_demand_checkpoint_requests != NULL);

		int i = 0;
		for(; i < (*desc)->initial_user_size; i++) {
			(*cp_desc)->queued_demand_checkpoint_requests[i] = gdsl_queue_alloc(NULL, NULL, NULL);
		}
	}
}

//#######################################################################################################################################
//routines for initializing demand checkpointing
//#######################################################################################################################################

//TODO: add the notification table (which procs we comm) to the checkpoint data
void init_checkpoint_window(MPI_Aint app_data_size) { 
	int cm_procs = (*desc)->cm_size;

	//First, we init the coordinated checkpoint window

	//TODO:
	// for now CC and D are integrated together; make them different

	//Then, we set up the demand checkpoint window

	int epoch_data_size = sizeof(int64_t) * cm_procs;// + sizeof(int64_t) * cm_procs;

	/* the final size of the whole checkpoint window */
	int cp_window_size = app_data_size * 2 + epoch_data_size * 2;

	assert(epoch_data_size % sizeof(uint64_t) == 0);
	assert(app_data_size % sizeof(uint64_t) == 0);

	int cp_disp_unit = sizeof(uint64_t);

	(*desc)->cp_data = (void*) malloc( cp_window_size );
	assert((*desc)->cp_data != NULL);

	/* we zero also the space occupied by the recorded epoch counters */
	memset((*desc)->cp_data, 0, cp_window_size);

	(*desc)->cp_win = malloc( sizeof(foMPI_Win) );
	assert((*desc)->cp_win != NULL);

	(*desc)->cp_data_disp_unit = cp_disp_unit;

	foMPI_Win_create( (*desc)->cp_data, cp_window_size, cp_disp_unit, MPI_INFO_NULL, (*desc)->initial_user_comm, (*desc)->cp_win );

	(*desc)->app_data_offset = 0;
	(*desc)->epoch_data_offset = app_data_size / cp_disp_unit;
	(*desc)->xor_app_data_offset = (app_data_size + epoch_data_size) / cp_disp_unit;
	(*desc)->xor_epoch_data_offset = (2 * app_data_size + epoch_data_size) / cp_disp_unit;

	(*desc)->data_offset = 0;
	(*desc)->xor_data_offset = (app_data_size + epoch_data_size) / cp_disp_unit;

	(*desc)->app_data_size = app_data_size;
	(*desc)->epoch_data_size = epoch_data_size;
}

//#######################################################################################################################################
//routines for initialize CM-specific windows and structures (epochs, logging spaces etc)
//#######################################################################################################################################

void init_cm_windows(foMPI_Win* app_win) {
	assert( *app_win != NULL );
	assert( (*app_win)->win_array != NULL );
	//assert( ((*app_win)->win_array)->base != NULL );

	(*cm_desc)->app_win = app_win;
	//TODO: double check if we need it if no logging is enabled
	init_epoch_window();
	debug_message_e( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, ">>> Epoch window initialized\n");

	if(do_logging()) {
		init_logging_space();
		debug_message_e( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, ">>> Logging space initialized\n");
	}

	init_marker_windows();
	debug_message_e( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, ">>> Marking windows initialized\n");

	init_counters();
	debug_message_e( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, ">>> Counters initialized\n");
}

void init_epoch_window() {
	// epoch window is shared only among CM processes 
	int cm_procs = (*desc)->cm_size;

	(*cm_desc)->epoch_data = (int64_t*) calloc( cm_procs, sizeof(int64_t) );
	assert((*cm_desc)->epoch_data != NULL);

	(*cm_desc)->epoch_win = malloc( sizeof(foMPI_Win) );
	assert((*cm_desc)->epoch_win != NULL);

	(*cm_desc)->local_epoch_counter = (int64_t*) calloc( cm_procs, sizeof(int64_t) );
	assert((*cm_desc)->local_epoch_counter != NULL);

	foMPI_Win_create((*cm_desc)->epoch_data, cm_procs * sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, *(*desc)->flib_level_comm, (*cm_desc)->epoch_win);
}

void init_logging_space() {
	(*cm_desc)->log_space = (Log_space*) malloc( sizeof(Log_space) );
	assert((*cm_desc)->log_space != NULL);

	Log_space* log_space = (*cm_desc)->log_space;

	log_space->local_list = malloc( sizeof(foMPI_Win) );
	log_space->global_list = malloc( sizeof(foMPI_Win) );

	assert(log_space->local_list != NULL);
	assert(log_space->global_list != NULL);

	size_t logging_list_size = log_list_size();

	char* local_list_data = (char*) malloc( logging_list_size * sizeof(char) );
	char* global_list_data = (char*) malloc( logging_list_size * sizeof(char) );
	assert(local_list_data != NULL);	
	assert(global_list_data != NULL);

	memset(local_list_data, 0, logging_list_size * sizeof(char) );
	memset(global_list_data, 0, logging_list_size * sizeof(char) );

	log_space->local_list_data = local_list_data;
	log_space->global_list_data = global_list_data;

	log_space->next_free_local_offset = 0;
	log_space->number_of_nodes_local = 0;
	log_space->local_sums = malloc( sizeof(int64_t) * (*desc)->cm_size );
	assert(log_space->local_sums != NULL);
	memset(log_space->local_sums, 0, sizeof(int64_t) * (*desc)->cm_size);

	log_space->__cleared_global_logs = 0;
	log_space->__cleared_global_test_logs = 0;

	log_space->global_logs_control_data = malloc( sizeof(int64_t) * ( (*desc)->cm_size + 2) );
	assert(log_space->global_logs_control_data != NULL);
	memset(log_space->global_logs_control_data, 0, sizeof(int64_t) * ( (*desc)->cm_size + 2));

	//init demand requests queues
	//TODO: change the size
	//TODO: double check (previous size: initial)
	log_space->gets_to_be_logged = malloc((*desc)->cm_size * sizeof(gdsl_queue_t));
	assert(log_space->gets_to_be_logged != NULL);

	//TODO: double check (previous size: initial)
	int i = 0;
	for(; i < (*desc)->cm_size; i++) {
		log_space->gets_to_be_logged[i] = gdsl_queue_alloc(NULL, NULL, NULL);
	}

	log_space->global_logs_control_win = malloc( sizeof(foMPI_Win) );
	assert(log_space->global_logs_control_win != NULL);

	log_space->puts_win = malloc( sizeof(foMPI_Win) );
	assert(log_space->puts_win != NULL);

	log_space->procs_communicated_using_puts = malloc( sizeof(int64_t) * (*desc)->cm_size );
	assert(log_space->procs_communicated_using_puts != NULL);

	log_space->procs_communicated_using_gets = malloc( sizeof(int64_t) * (*desc)->cm_size );
	assert(log_space->procs_communicated_using_gets != NULL);

	for(i = 0; i < (*desc)->cm_size; i++) {
		(log_space->procs_communicated_using_puts)[i] = FALSE;
		(log_space->procs_communicated_using_gets)[i] = FALSE;
	}

	//TODO: reduce the number of windows
	log_space->local_list_control = malloc( sizeof(foMPI_Win) );
	assert(log_space->local_list_control != NULL);

	log_space->local_list_size = malloc( sizeof(foMPI_Win) );
	assert(log_space->local_list_size != NULL);

	foMPI_Win_create(&(log_space->number_of_nodes_local), sizeof(int), sizeof(int), MPI_INFO_NULL, *(*desc)->flib_level_comm, log_space->local_list_control); 
	foMPI_Win_create(&(log_space->next_free_local_offset), sizeof(int), sizeof(int), MPI_INFO_NULL, *(*desc)->flib_level_comm, log_space->local_list_size); 

	foMPI_Win_create(log_space->local_list_data, logging_list_size * sizeof(char), sizeof(char), MPI_INFO_NULL, *(*desc)->flib_level_comm, log_space->local_list); 
	foMPI_Win_create(log_space->global_list_data, logging_list_size * sizeof(char), sizeof(char), MPI_INFO_NULL, *(*desc)->flib_level_comm, log_space->global_list); 

	foMPI_Win_create(log_space->global_logs_control_data, ( (*desc)->cm_size + 2) * sizeof(int64_t), sizeof(int64_t),
			MPI_INFO_NULL, *(*desc)->flib_level_comm, log_space->global_logs_control_win);	
	foMPI_Win_create(log_space->procs_communicated_using_gets, ( (*desc)->cm_size) * sizeof(int64_t), sizeof(int64_t),
			MPI_INFO_NULL, *(*desc)->flib_level_comm, log_space->puts_win);
}

void init_marker_windows() {	
	(*cm_desc)->marker_M_win = malloc( sizeof(foMPI_Win) );
	assert((*cm_desc)->marker_M_win != NULL);

	(*cm_desc)->marker_M_space = malloc((*desc)->cm_size * sizeof(int));
	assert((*cm_desc)->marker_M_space != NULL);

	foMPI_Win_create((*cm_desc)->marker_M_space, ( (*desc)->cm_size) * sizeof(int), sizeof(int),
			MPI_INFO_NULL, *(*desc)->flib_level_comm, (*cm_desc)->marker_M_win);
}

void init_counters() {
	(*cm_desc)->GNC_cnt = 0;
  (*cm_desc)->SC_cnt = 0;
  (*cm_desc)->GC_cnt = 0;

	//TODO: develop and init the window for the SC counter
}
	
//#######################################################################################################################################
//take an initial checkpoint
//#######################################################################################################################################

void take_initial_checkpoint() {
	// in the initial checkpoint, we do not care about recording epoch counters, as all procs are in epoch zero 
	MPI_Barrier((*desc)->initial_user_comm);

	MPI_Status s;

	uint64_t* checkpoint_data = &((uint64_t*)(*desc)->cp_data)[(*desc)->data_offset];
	uint64_t* xored_data = &((uint64_t*)(*desc)->cp_data)[(*desc)->xor_data_offset];

	if(cm_proc()) {

		/* IMPORTANT: based on Robert's implementaton */
		uint64_t* app_data = (uint64_t*) ((*((*cm_desc)->app_win))->win_array)[(*desc)->flib_level_rank].base;
		int app_data_size = (*desc)->app_data_size;
		int cp_data_unit = (*desc)->cp_data_disp_unit;

		int assigned_proc = assigned_cp_proc((*desc)->initial_user_rank);

		MPI_Send(app_data, app_data_size / cp_data_unit, MPI_UINT64_T, assigned_proc, CM_INITIAL_CHECKPOINT, (*desc)->initial_user_comm);

		/* in the initial checkpoint, we copy the application data to both parts of the checkpoint window */
		memcpy(checkpoint_data, app_data, app_data_size);
		memcpy(xored_data, app_data, app_data_size);
	}
	else {
		int* assigned_procs = assigned_cm_procs();
		int assigned_procs_nr = assigned_cm_procs_nr();
		int app_data_size = (*desc)->app_data_size;
		int cp_data_unit = (*desc)->cp_data_disp_unit;

		/* we receive the checkpoint data from every assigned CM process */
		while(assigned_procs_nr > 0) {
			MPI_Recv(checkpoint_data, app_data_size / cp_data_unit, MPI_UINT64_T, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);

			assert(s.MPI_TAG == CM_INITIAL_CHECKPOINT);
			assert(is_cm_proc(s.MPI_SOURCE));

			/* we make sure that the source is our CM process */
			assert(is_our_proc( s.MPI_SOURCE, _FLIB_INITIAL_COMM ) == TRUE);

			/* upon the first message, we just copy the data. */
			/* after the first message, we perform the actual xoring*/
			if(assigned_procs_nr == assigned_cm_procs_nr() ) {
				memcpy(xored_data, checkpoint_data, app_data_size);
			}
			else {
				calculate_xor(checkpoint_data, xored_data, app_data_size);
			}
			assigned_procs_nr--;
		}

		free(assigned_procs);
	}

	MPI_Barrier((*desc)->initial_user_comm);

#ifdef FLIB_DEBUG
	__test_initial_checkpoint();
#endif
}
