#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#include "fompi.h"
#include "flib.h"
#include "common.h"
#include <assert.h>

// the size of the whole hashtable
int64_t TSIZE;

//FLIB variables
MPI_Comm flib_level_comm;
int cp_procs;

int hierarchy;
float mtbf;
int protocol_type;

double start_time;
double time_to_last;
//end of FLIB variables

int local_comp;

//variables used for testing if the results are correct
int insert_collisions = 0;
int delete_collisions = 0;
int64_t inserts = 0;
int deletes = 0;

/*
	 in this version, we use only int64_t elements
	 an element in table/heap is a pair of int64_t variables.
	 the first such variable is the element
	 the next variable is the 'next'
 */

typedef struct {
	foMPI_Win win;
	int64_t *table; // the table/heap + last + next_free
	int p, r;
	int commsize, commrank;
	int64_t tsize;  // size of table (in elements)
	int64_t size;   // size of the heap (in elements)

	int64_t t_offset;	//offset of the table/heap space from the beginning of the window
	int64_t l_offset;	//offset of the last space from the beginning of the window
	int64_t n_offset; //offset of the next free space from the beginning of the window

	int64_t table_size;	//size of table/heap (in int64_t)
	int64_t last_size;	//size of last array (in int64_t)
	int64_t nextfree_size;	//size of nextfree (in int64_t)
} t_hash;

int64_t hashfunc(int64_t val) {
	return val & (TSIZE-1);
}

void printHTLocal(t_hash* hash) {
	printf("[%i] @@@@@@ Table: \n ",hash->r);
	for(int i = 0; i < hash->tsize; i++) {    
		printf("[%i] %i : (value: %i, next: %i)\n",hash->r,i,(hash->table)[i*2], (hash->table)[i*2+1]);
	}

	printf("\n[%i] @@@ Heap: \n",hash->r);
	for(int i = hash->tsize; i < hash->tsize+hash->size; i++) {
		printf("[%i] %i : (value: %i, next: %i)\n",hash->r,i,(hash->table)[2*i], (hash->table)[2*i+1]); 
	}
}

void insert(t_hash* hash, int64_t elem) {
	int64_t global_pos = hashfunc(elem);
	int64_t owner = global_pos/hash->tsize;

	// CAS if table position is available
	int64_t compare = -1;
	int64_t result = -2;
	int64_t local_pos = global_pos - owner*hash->tsize;
	int64_t offset /* in ints */ = hash->t_offset + local_pos*2;

#ifdef HASHTABLE_DEBUG
	fprintf(stderr,"[%i] insert %lld (local pos: %lld, pos: %lld) - owner %lld\n", hash->r, elem, offset / 2, global_pos, owner);
#endif

	fMPI_Compare_and_swap(&elem, &compare, &result, MPI_INT64_T, owner, /* target disp */ offset, hash->win);
	fMPI_Win_flush(owner, hash->win);

	if(result != compare) {
		++insert_collisions;

		int64_t origin = 1; // add 1
		int64_t result = -1; // address of newly added record

		// grab remote location
		fMPI_Fetch_and_op(/* origin */ &origin, /* result */ &result, MPI_INT64_T, owner, /* disp */ hash->n_offset, foMPI_SUM, hash->win);
		fMPI_Win_flush(owner, hash->win);

		if(result >= hash->tsize+hash->size) {
			fprintf(stderr,"Failed to insert due to the OVERFLOW\n");
			return;
		}

		// put element in grabbed location

		fMPI_Put(&elem, 1, MPI_INT64_T, owner, hash->t_offset + 2*(result), 1, MPI_INT64_T, hash->win);

		int64_t newOffset = result;		
		int64_t oldOffset = -111;  

		// grab lastptr element and plug in ptr to new element
		fMPI_Fetch_and_op(/* origin */ &newOffset, /* result */ &oldOffset, MPI_INT64_T, owner, /* disp */ hash->l_offset + local_pos, foMPI_REPLACE, hash->win);
		fMPI_Win_flush(owner, hash->win);

		compare = -1;
		result = -2;
		offset /* in ints */ = local_pos*2+1;

		// try to update table pointer directly (if it's the first update)
		fMPI_Compare_and_swap(&newOffset, &compare, &result, MPI_INT64_T, owner, /* target disp */ hash->t_offset + offset, hash->win);
		fMPI_Win_flush(owner, hash->win);

		if(result != compare) { // we lost the CAS for the direct pointer -- no direct access to table!
			fMPI_Put(&newOffset, 1, MPI_INT64_T, owner, hash->t_offset + 2*(oldOffset)+1, 1, MPI_INT64_T, hash->win);
			fMPI_Win_flush(owner, hash->win);
		}
	}
}

//count the number of elements by following the pointers in the hashtable
int64_t count(t_hash *hash) {
	int64_t cnt=0;
	int64_t heapcnt=0;
	for(int i=0; i<hash->tsize; i++) {
		int pos = i;
		if(hash->table[2*pos] != -1) {
			cnt++;
			if(hash->table[2*pos+1] != -1) {
				heapcnt++;
				pos=hash->table[2*pos+1];
				while(hash->table[2*pos+1] != -1) {
					pos = hash->table[2*pos+1];
					heapcnt++;
				}
			}
		}
	}
	return cnt+heapcnt;
}

//count by iterating the whole table+heap, do not follow the pointers
int64_t countSimple(t_hash* hash) {
	int64_t cnt = 0;
	for(int i = 0; i < (hash->tsize+hash->size)*2;i+=2) {
		if(hash->table[i] != -1) {
			cnt++;
		}
	}
	return cnt;
}

void clear_hashtable(t_hash* hash) {
	hash->table[hash->n_offset] = hash->tsize; // next free element in heap

	for(int i=0; i<(hash->tsize+hash->size)*2; i+=2) {
		hash->table[hash->t_offset + i] = -1;
		hash->table[hash->t_offset + i + 1] = -1;
	}

	for(int i=0; i < hash->tsize; i++) {
		hash->table[hash->l_offset + i] = -1;
	}
}

int main(int argc, char **argv) {
	foMPI_Init(&argc, &argv);

	t_hash hash;
	int LV = atoi(argv[1]); //TODO: it's an integer!!!

	cp_procs = atoi(argv[2]);
	hierarchy = atoi(argv[3]);
	mtbf = atof(argv[4]);
	protocol_type = atoi(argv[5]);
	time_to_last = atof(argv[6]);
	local_comp = atoi(argv[7]);
	int log_size = atoi(argv[8]); 
	iters = atoi(argv[9]); int puts = atoi(argv[10]); int gets = atoi(argv[11]);


	assert(protocol_type == FLIB_NO_COORDINATED_SCHEME);

	MPI_Comm_size(MPI_COMM_WORLD, &hash.commsize);
	MPI_Comm_rank(MPI_COMM_WORLD, &hash.commrank);

	if(hash.commrank == 0) {
	//	printf(">>> CP processes: %i, hierarchy: %i, mtbf: %f, protocol_type: %i\n", cp_procs, hierarchy, mtbf, protocol_type);
	}

	comp_time = 0;

	hash.tsize = LV;
	hash.size = (MAX_AVG_COLLISIONS)*LV;

	hash.table_size = (hash.tsize + hash.size)*sizeof(int64_t)*2;
	hash.last_size = hash.tsize * sizeof(int64_t);
	hash.nextfree_size = sizeof(int64_t);

	hash.t_offset = 0;
	hash.l_offset = (hash.tsize + hash.size)*2;
	hash.n_offset = hash.l_offset + hash.tsize;	

	if(hash.commrank == 0) {
		fprintf(stderr,"[%i] Hashtable offsets calculated\nITERS: %i\n", hash.r, iters);
	}

	fMPI_Win_allocate(hash.table_size + hash.last_size + hash.nextfree_size, sizeof(int64_t), MPI_INFO_NULL, MPI_COMM_WORLD, &hash.table, 
			&hash.win, cp_procs, &flib_level_comm, hierarchy, mtbf, protocol_type, puts, gets, log_size);

	if(cm_proc()) 
	{
	if(hash.commrank == 0) {
		fprintf(stderr,"[%i] Hashtable WIN allocated\n", hash.r);
	}

		if(protocol_type != FLIB_NO_COORDINATED_SCHEME) {
			double daly_int = fLIB_get_daly_interval();

			//int check_freq = (int)(time_to_last / daly_int) + 1;
			//set_stopping_frequency(check_freq);
		}

		hash.p = fLIB_get_user_size();
		hash.r = fLIB_get_user_rank();

		TSIZE = (int64_t)LV * (int64_t)hash.p;
		init_genrand64(hash.r*1000000);
		//srand(hash.r*1000000);
		//assert(RAND_MAX > TSIZE);
		assert(hash.commsize > hash.p);

		HRT_INIT(1, g_timerfreq);

		clear_hashtable(&hash);

#ifdef HASHTABLE_DEBUG
		if(hash.r == 0) printHTLocal(&hash);
#endif

		int nr_of_ins = HTABLE_FILL_RATIO*LV;

		MPI_Barrier(flib_level_comm);

		int iterations = 0;
		if(hash.r == 0) {
			fprintf(stderr,">>> benchmark START\n");
		}
		start_time = MPI_Wtime();

		while(1) {
			int64_t val = genrand64_int63(); 
			insert(&hash, val);
			inserts++;
			if(local_comp) {
				local_computation();
			}

			/*if(protocol_type != FLIB_NO_COORDINATED_SCHEME) {
				if(hash.r == 0) {
				double current_time = MPI_Wtime();

				if(current_time - start_time >= time_to_last) {
				stop_coordinated_checkpointing();
				}
				}

				if(coordinated_checkpointing_stopped()) {
				break;
				}
				}
				else*/ {
					double current_time = MPI_Wtime();

					if(current_time - start_time >= time_to_last) {
						break;
					}
				}
				iterations++;
		}

		MPI_Barrier(flib_level_comm);

		double stop_time = MPI_Wtime();
		int64_t inserts_sum = 0;

		MPI_Reduce(&inserts, &inserts_sum, 1, MPI_INT64_T, MPI_SUM, 0, flib_level_comm);

		if(hash.r == 0) {
			double ins_time = stop_time-start_time;
			//ins_time /= 1.0e6;
			comp_time /= 1.0e6;

			double inserts_per_sec = ((double)inserts_sum/ins_time);
			double comp_to_comm = (comp_time/ins_time)*100;
			//double colls_ratio = 100* (double)collisions_sum / ((double)inserts_to_do * (double)hash.p);
			//int64_t all_ins = hash.p*(inserts_to_do - inserts_to_warmup);

			printf("%i\t%f\n",hash.p,inserts_per_sec);
			fprintf(stdout,"procs: %i, ins: %lli, comp_time: %f, all_time: %f, comp/comm: %f\n",
			          hash.p, inserts_sum, comp_time, ins_time, comp_to_comm);
			//fprintf(stdout,"procs: %i, ins: %lli, time: %f, comp/comm: %f\n",
		//			hash.p, inserts_sum, ins_time, comp_to_comm);

			//fprintf(stderr,">>> iteration %i END\n", iterations);
			//printf("%i %f\n",hash.p,inserts_sum / (stop_time - start_time));
		}

#ifdef HASHTABLE_DEBUG
		for(int i = 0; i < hash.p; i++) {
			if(hash.r == i) printHTLocal(&hash);
			MPI_Barrier(flib_level_comm);
		}
#endif

		fprintf(stderr,"[%i] collisions: %i\n",hash.r,insert_collisions );

		int64_t size = count(&hash);

		//	#ifdef HASHTABLE_DEBUG
		fprintf(stderr,"[%i] cnt: %i, simpleCnt:%lld\n",hash.r,size,countSimple(&hash));
		//	#endif

		assert(size == countSimple(&hash));

		MPI_Barrier(flib_level_comm);

		if(hash.r == 0) fprintf(stderr,">>> TEST 1 (sum and simple sum) OK\n");

		int64_t sum = 0;
		MPI_Reduce(&size, &sum, 1, MPI_INT64_T, MPI_SUM, 0, flib_level_comm);

		if(hash.r == 0) {
			fprintf(stderr,"[%i] inserts_sum: %i, sum:%lld\n",hash.r,inserts_sum,sum);
			assert(inserts_sum == sum);
			fprintf(stderr,"TEST 2 (The sum of inserts: %lld) OK\n", sum);
		}

		MPI_Barrier(flib_level_comm);
	}

	fMPI_Win_free(&hash.win);
	foMPI_Finalize();

	return 0;
}
