#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#include "fompi.h"
#include "common.h"
#include <assert.h>

// the size of the whole hashtable
int64_t TSIZE;

double start_time;
double time_to_last;

//variables used for testing if the results are correct
int64_t insert_collisions = 0;
int delete_collisions = 0;
int64_t inserts = 0;
int deletes = 0;

/*extern "C" {
  typedef struct{
    int64_t value;
    int64_t next;
  } t_elem;
}*/

/*
	in this version, we use only int64_t elements
	an element in table/heap is a pair of int64_t variables.
	the first such variable is the element
	the next variable is the 'next'
*/

typedef struct {
  foMPI_Win win;
  int64_t *table; // the table/heap + last + next_free
  int p, r;
  int64_t tsize;  // size of table (in elements)
  int64_t size;   // size of the heap (in elements)

	int64_t t_offset;	//offset of the table/heap space from the beginning of the window
	int64_t l_offset;	//offset of the last space from the beginning of the window
	int64_t n_offset; //offset of the next free space from the beginning of the window

	int64_t table_size;	//size of table/heap (in int64_t)
	int64_t last_size;	//size of last array (in int64_t)
	int64_t nextfree_size;	//size of nextfree (in int64_t)
} t_hash;

int64_t hashfunc(int64_t val) { 
        return val & (int64_t)(TSIZE-(int64_t)1);
}

void printHTLocal(t_hash* hash) {
  printf("[%i] @@@@@@ Table: \n ",hash->r);
  for(int i = 0; i < hash->tsize; i++) {    
    printf("[%i] %i : (value: %i, next: %i)\n",hash->r,i,(hash->table)[i*2], (hash->table)[i*2+1]);
  }

  printf("\n[%i] @@@ Heap: \n",hash->r);
  for(int i = hash->tsize; i < hash->tsize+hash->size; i++) {
    printf("[%i] %i : (value: %i, next: %i)\n",hash->r,i,(hash->table)[2*i], (hash->table)[2*i+1]); 
  }
}

void insert(t_hash* hash, int64_t elem) {
  int64_t global_pos = hashfunc(elem);
  int64_t owner = global_pos/hash->tsize;
	
	// CAS if table position is available
	int64_t compare = -1;
	int64_t result = -2;
	int64_t local_pos = global_pos - owner*hash->tsize;
	int64_t offset /* in ints */ = hash->t_offset + local_pos*2;

#ifdef HASHTABLE_DEBUG
	fprintf(stderr,"[%i] insert %lld (local pos: %lld, pos: %lld) - owner %lld\n", hash->r, elem, offset / 2, global_pos, owner);
#endif

	foMPI_Compare_and_swap(&elem, &compare, &result, MPI_INT64_T, owner, /* target disp */ offset, hash->win);
	foMPI_Win_flush_local(owner, hash->win);
	
	if(result != compare) {
		++insert_collisions;

		int64_t origin = 1; // add 1
		int64_t result = -1; // address of newly added record

		// grab remote location
		foMPI_Fetch_and_op(/* origin */ &origin, /* result */ &result, MPI_INT64_T, owner, /* disp */ hash->n_offset, foMPI_SUM, hash->win);
		foMPI_Win_flush_local(owner, hash->win);
		
		if(result >= hash->tsize+hash->size) {
			fprintf(stderr,"Failed to insert due to the OVERFLOW\n");
			return;
		}

		// put element in grabbed location
				
		foMPI_Put(&elem, 1, MPI_INT64_T, owner, hash->t_offset + 2*(result), 1, MPI_INT64_T, hash->win);

		int64_t newOffset = result;		
		int64_t oldOffset = -111;  

		// grab lastptr element and plug in ptr to new element
		foMPI_Fetch_and_op(/* origin */ &newOffset, /* result */ &oldOffset, MPI_INT64_T, owner, /* disp */ hash->l_offset + local_pos, foMPI_REPLACE, hash->win);
		foMPI_Win_flush_local(owner, hash->win);

		compare = -1;
		result = -2;
		offset /* in ints */ = local_pos*2+1;

		// try to update table pointer directly (if it's the first update)
		foMPI_Compare_and_swap(&newOffset, &compare, &result, MPI_INT64_T, owner, /* target disp */ hash->t_offset + offset, hash->win);
		foMPI_Win_flush_local(owner, hash->win);

		if(result != compare) { // we lost the CAS for the direct pointer -- no direct access to table!
			foMPI_Put(&newOffset, 1, MPI_INT64_T, owner, hash->t_offset + 2*(oldOffset)+1, 1, MPI_INT64_T, hash->win);
			foMPI_Win_flush(owner, hash->win);
		}
	}
}

//count the number of elements by following the pointers in the hashtable
int64_t count(t_hash *hash) {
  int64_t cnt=0;
  int64_t heapcnt=0;
  for(int i=0; i<hash->tsize; i++) {
    int pos = i;
    if(hash->table[2*pos] != -1) {
      cnt++;
      if(hash->table[2*pos+1] != -1) {
        heapcnt++;
        pos=hash->table[2*pos+1];
        while(hash->table[2*pos+1] != -1) {
          pos = hash->table[2*pos+1];
          heapcnt++;
        }
      }
    }
  }
  return cnt+heapcnt;
}

//count by iterating the whole table+heap, do not follow the pointers
int64_t countSimple(t_hash* hash) {
  int64_t cnt = 0;
  for(int i = 0; i < (hash->tsize+hash->size)*2;i+=2) {
    if(hash->table[i] != -1) {
      cnt++;
    }
  }
  return cnt;
}

void clear_hashtable(t_hash* hash) {
  hash->table[hash->n_offset] = hash->tsize; // next free element in heap

  for(int i=0; i<(hash->tsize+hash->size)*2; i+=2) {
    hash->table[hash->t_offset + i] = -1;
    hash->table[hash->t_offset + i + 1] = -1;
  }

  for(int i=0; i < hash->tsize; i++) {
    hash->table[hash->l_offset + i] = -1;
  }
}

int main(int argc, char **argv) {
  foMPI_Init(&argc, &argv);
  
  t_hash hash;
  int LV = atoi(argv[1]);
	time_to_last = atof(argv[2]); 
 
  MPI_Comm_size(MPI_COMM_WORLD, &hash.p);
  MPI_Comm_rank(MPI_COMM_WORLD, &hash.r);
  
  TSIZE = (int64_t)LV * (int64_t)hash.p;
 	srand(hash.r*1000000);
  init_genrand64(hash.r*1000000);
  //assert(RAND_MAX > TSIZE);
  
  hash.tsize = LV;
  hash.size = (MAX_AVG_COLLISIONS)*LV;

	hash.table_size = (hash.tsize + hash.size)*sizeof(int64_t)*2;
	hash.last_size = hash.tsize * sizeof(int64_t);
	hash.nextfree_size = sizeof(int64_t);

	hash.t_offset = 0;
	hash.l_offset = (hash.tsize + hash.size)*2;
	hash.n_offset = hash.l_offset + hash.tsize;	

#ifdef HASHTABLE_DEBUG
	 fprintf(stderr,"[%i] Hashtable offsets calculated\n", hash.r);
#endif

  foMPI_Win_allocate(hash.table_size + hash.last_size + hash.nextfree_size, sizeof(int64_t), MPI_INFO_NULL, MPI_COMM_WORLD, &hash.table, &hash.win);

#ifdef HASHTABLE_DEBUG
	fprintf(stderr,"[%i] Hashtable WIN allocated\n", hash.r);
#endif

	clear_hashtable(&hash);

#ifdef HASHTABLE_DEBUG
	if(hash.r == 0) printHTLocal(&hash);
	fprintf(stderr,"[%i] Hashtable initialized\n", hash.r);
#endif

	int nr_of_ins = HTABLE_FILL_RATIO*LV;

	MPI_Barrier(MPI_COMM_WORLD);

	int iterations = 0;
	if(hash.r == 0) {
		fprintf(stderr,">>> benchmark START\n");
		start_time = MPI_Wtime();
	}
  start_time = MPI_Wtime();


	while(1) {
		//for(int m = 0; m < nr_of_ins; m++) {
			int64_t val = genrand64_int63();
			insert(&hash, val);
			inserts++;
			local_computation();
		//}

		double current_time = MPI_Wtime();

		if(current_time - start_time >= time_to_last) {
			break;
		}
		iterations++;
	}

  MPI_Barrier(MPI_COMM_WORLD);

	double stop_time = MPI_Wtime();
	int64_t inserts_sum = 0;
	int64_t insert_collision_sum = 0;

	MPI_Reduce(&inserts, &inserts_sum, 1, MPI_INT64_T, MPI_SUM, 0, MPI_COMM_WORLD);
	MPI_Reduce(&insert_collisions, &insert_collision_sum, 1, MPI_INT64_T, MPI_SUM, 0, MPI_COMM_WORLD);

#ifdef HASHTABLE_DEBUG
	for(int i = 0; i < hash.p; i++) {
		if(hash.r == i) printHTLocal(&hash);
		MPI_Barrier(MPI_COMM_WORLD);
	}
#endif

	if(hash.r == 0) {
		fprintf(stderr,">>> iteration %i END. Time: %f\n", iterations, (double)(stop_time - start_time));
		printf("%i %f\n",hash.p,(double)inserts_sum / (double)(stop_time - start_time)); //ins/sec
	  fprintf(stderr,"[%i] inserts: %lli\n",hash.r,inserts_sum);
	  fprintf(stderr,"[%i] collisions: %lli\n",hash.r,insert_collision_sum );
	  fprintf(stderr,"[%i] collisions (percentage of all inserts): %f\n",hash.r,((double)insert_collision_sum/(double)inserts_sum) * 100 );
	}

  int64_t size = count(&hash);

#ifdef HASHTABLE_DEBUG
  fprintf("[%i] cnt: %lld, simpleCnt:%lld\n",hash.r,size,countSimple(&hash));
#endif

  assert(size == countSimple(&hash));

  MPI_Barrier(MPI_COMM_WORLD);

	if(hash.r == 0) fprintf(stderr,">>> TEST 1 (sum and simple sum) OK\n");

  int64_t sum = 0;
  MPI_Reduce(&size, &sum, 1, MPI_INT64_T, MPI_SUM, 0, MPI_COMM_WORLD);

   if(hash.r == 0) {
    assert(inserts_sum == sum);
    fprintf(stderr,"TEST 2 (The sumof inserts: %lld) OK\n", sum);
  }

  foMPI_Finalize();

	return 0;
}
